/**
	@file 	pGina.c
	@brief 	pGina support functions.
*/
/*
pGina - A modular GINA Interface for Windows
Copyright (C) 2002 Nathan Yocom, Michael Wright

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email: nate.yocom@xpasystems.com
Web: http://pgina.xpasystems.com
Snail Mail:
  Nathan Yocom 
  9 Evergreen Farms Rd.
  Scarborough, ME 04074
  Phone: 207-450-4948
*/

#include <windows.h>
#include <winwlx.h>
#include <stdio.h>
#include <Winreg.h>
#include <DsGetDC.h>
#include "pGina.h"
#include "resource.h"
#include <Userenv.h>

#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

// For security-related API calls
#include <aclapi.h>
#include <sddl.h>

/**
  @brief Called when the pGina DLL is first attached by WinLogon (or other program)
  
  Called when the pGina DLL is first attached by WinLogon (or other program)
  we save the hInstance for later use in mapping to other functions etc.  This
  is also where we load the encryption dll (if turned on), and select the appropriate
  language for displaying text in the dialogs.
  
  @param 	hInstance	The instance of this particular DLL loading
  @param	dwReason	The reason for being attached/detached
  @param	lpReserved	Reserved by Microsoft (c)   
  @retval 	boolean     Unless there is a major problem, return true
*/
BOOL WINAPI DllMain(HINSTANCE hInstance,DWORD dwReason,LPVOID lpReserved)
{
	switch(dwReason)					  // Decide why we have entered 
	{
	case DLL_PROCESS_ATTACH:			  // If we entered and are being attached to 
		pGinahInstance = hInstance;		  // Save our parents hInstance

        // If encryptSettings is turned on, load the encryption Key DLL
		encryptSettings = regReadBool(_T("encryptSettings")); 
		if(encryptSettings) {
            GetEncryptionKey();
		}
        idleTime = 30;
        LoadSettings();	  // Load configuration from reg

        getLocaleFile();

        srand(time(NULL));                // Seed the crappy prng

        hDll = NULL;
		debugFile = NULL;

        windowsVersion = getWindowsVersion();
		defaultSessionTimeout = 0;
		useSLoginCredentials = false;
        desktopLocked = false;
        timedOutLogoff = false;
        idleLockHit = false;
        loginAttemptCount = 0;
   	case DLL_PROCESS_DETACH:
    default: 
		return(TRUE);
	}
}

/**
  @brief 	Configure pGina settings

	The pGinaConfigure() method is called by the external configuration application.  
	This method is responsible for providing a GUI interface and functionality for 
	the configuration of the pGina DLL and its settings.  Settings include all members
	of the pGinaInfo structure except the username and password as well as all other
    global values.

  @retval boolean
	In the event of any errors, this method will provide all necessary warnings and 
	information to the user prior to returning a false value, otherwise a value of true
	will be returned.

  @see ShowConfigDialog
*/
EXTERNAL_API BOOL pGinaConfigure(void)
{
	return ShowConfigDialog();
}

/**
	@brief 	Load a pGina plugin and map functions within it
	@author Nathan Yocom
	
	Called to map all the necessary methods from the plugin.
  
	@retval	boolean indicating success or failure
 
	In the event of any errors, this method will provide all necessary warnings and 
	information to the user prior to returning a false value, otherwise a value of true
	will be returned.
*/
BOOL LoadPlugin(void)
{
	if(hDll != NULL)
	{
		pluginUserLogin = NULL;
		pluginChangeUserPassword = NULL;
		pluginAboutPlugin = NULL;
		pluginChangePluginSettings = NULL;
		pluginLoginHook = NULL;
		pluginLogoutHook = NULL;
		pluginIsRequired = NULL;
		FreeLibrary(hDll);
	}

	// Make sure its a DLL and set the handle
	if(!(hDll = LoadLibrary(settingsInfo.pathPlugin)))
	{
		// Notify the user if there were issues
		// "This does not appear to be a valid plugin."
		//pGinaMessageBox(IDS_PLINVALID,(UINT)0,MB_OK);
		MessageBox(GetForegroundWindow(),TranslatedString(GSECTION,TEXT("IDS_PLINVALID")),NULL,MB_OK);
		EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Error occured in loading the plugin (LoadLibrary failed)."));
		return false;
	}
	else
	{
		// Map the methods
		pluginUserLogin = (DLLUSERLOGIN) GetProcAddress(hDll,"UserLogin");
		pluginChangeUserPassword = (DLLCHANGEUSERPASSWORD) GetProcAddress(hDll,"ChangeUserPassword");
		pluginAboutPlugin = (DLLABOUTPLUGIN) GetProcAddress(hDll,"AboutPlugin");
		pluginChangePluginSettings = (DLLCHANGEPLUGINSETTINGS) GetProcAddress(hDll,"ChangePluginSettings");
		pluginLoginHook = (DLLLOGINHOOK) GetProcAddress(hDll,"LoginHook");
		pluginLogoutHook = (DLLLOGOUTHOOK) GetProcAddress(hDll,"LogoutHook");
		pluginIsRequired = (DLLISREQUIRED) GetProcAddress(hDll,"IsRequired");

		// If any of them were NOT loaded
		if(!pluginUserLogin || !pluginChangeUserPassword || !pluginAboutPlugin || !pluginChangePluginSettings || !pluginLoginHook || !pluginLogoutHook)
		{
			// Tell the user something was missing/wrong
			MessageBox(GetForegroundWindow(),TranslatedString(GSECTION,TEXT("IDS_PLINVALIDINCOM")),TranslatedString(GSECTION,TEXT("IDS_T_LOADERR")),MB_OK);
			return false;
		}
		else
		{
			// If we get here things are cool...
			// Check for required status (must allow older plugins to still work (old ones are optional):
			if(!pluginIsRequired)
				isRequired = false;
			else
				isRequired = pluginIsRequired();

			return true;
		}
	}
	// if we get this far something is DEFINETLY wrong!
	return false; 
}

/**
	@brief	Map network resources to local logical drives
	
	The MapDrives() method is called to parse and map all drives as specified 
	in the pGinaInfo structure.

	@retval	boolean
	In the event of any errors, this method will provide all necessary warnings and 
	information to the user prior to returning a false value, otherwise a value of true
	will be returned.
*/
BOOL MapDrives(void)
{
	TCHAR *current;		// Current entry
	TCHAR *list;			// The full list
	TCHAR drive[3];		// The drive to map
	TCHAR *path;			// the path to map to *drive
	LONG lResult;
	BOOL errors = FALSE;	// Were there problems

	TCHAR *buffer = NULL;	// For expanding username
	DWORD bufsize = 0;

	NETRESOURCE NetSettings;  // The structure for mapping info
	path = NULL;
	NetSettings.dwType = RESOURCETYPE_DISK;
	NetSettings.lpProvider = NULL;
	NetSettings.dwDisplayType   = RESOURCEDISPLAYTYPE_SHARE;
	NetSettings.dwScope         = RESOURCE_GLOBALNET;
	NetSettings.dwUsage         = RESOURCEUSAGE_CONNECTABLE;
	NetSettings.lpComment       = NULL;

	if(settingsInfo.mapPaths == NULL)
		return true;

	// Moved to avoid 5 secs per mount, thanks Skindog!
	Sleep(5000);	// Takes milliseconds as argument

	// Copy the main list (we don't want to screw with that!)
	list = _tcsdup(settingsInfo.mapPaths);

	// Make current point to an expanded string
	buffer = (TCHAR *)malloc((wcslen(list) * 2) * sizeof(TCHAR) + 1);
	bufsize = (wcslen(list) * 2) + 1;
	memset(buffer,'\0',bufsize);

	// Always use our custom expander here, we have to replace %username% with the correct value
	//  depending on the single user login option
	ExpandEnvStrings(list,buffer,bufsize);

	// Get the first item...
	current = _tcstok(buffer,TEXT(";"));
	// Until there are no more items
	while(current != NULL)
	{
		// Current now has "DRIVE:\\Share\"
		if(path)
			free(path);

		// Path should be the first \ in the string
		// This was doing something very nasty (static +2) :)
		path = _tcsdup(wcsstr(current,TEXT("\\")));
		//_tcsncpy(drive,current,2);
		drive[0] = current[0];
		drive[1] = TEXT(':');
		drive[2] = TEXT('\0');

		NetSettings.lpLocalName = drive;
		NetSettings.lpRemoteName = path;

		EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Trying to map %s:%s"),drive,path);

		// All variables are there now - DO THE MAP
		if(!ImpersonateLoggedOnUser(hUser))
		{
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Impersonation during mapping failed."));
		}

		//EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Trying to map %s:%s with username: %s password: %s"),drive,path,settingsInfo->Username,settingsInfo->Password);
		lResult = WNetAddConnection2(&NetSettings,(LPCTSTR)settingsInfo.Password,(LPCTSTR) settingsInfo.Username,CONNECT_INTERACTIVE);

		if(lResult != NO_ERROR)
		{
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Mapping of drive (non-interactively) failed: %d"),lResult);
			lResult = WNetAddConnection2(&NetSettings,NULL,NULL,CONNECT_PROMPT | CONNECT_INTERACTIVE);
		}

		if(lResult != NO_ERROR)
		{
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Mapping of drive interactively failed also: %d"),lResult);

			if(path)
				free(path);

			if(list)
				free(list);

			errors = TRUE;
		}

		// Get next token
		current = _tcstok(NULL ,TEXT(";"));
	}

	// If we get this far, then all went well

	if(list)
		free(list);
	if(buffer)
		free(buffer);
	RevertToSelf();	// Reverse the impersonation

	return !errors;
}

void DeMapDrives(void)
{
	TCHAR *current;		// Current entry
	TCHAR *list;			// The full list
	TCHAR drive[3];		// The drive to map
	TCHAR *path;			// the path to map to *drive
	LONG lResult;

	TCHAR *buffer = NULL;	// For expanding username
	DWORD bufsize = 0;

    path = NULL;
	
	if(settingsInfo.mapPaths == NULL)
		return;

    // Copy the main list (we don't want to screw with that!)
	list = _tcsdup(settingsInfo.mapPaths);

	// Make current point to an expanded string
	buffer = (TCHAR *)malloc((wcslen(list) * 2) * sizeof(TCHAR) + 1);
	bufsize = (wcslen(list) * 2) + 1;
	memset(buffer,'\0',bufsize);

	// Always use our custom expander here, we have to replace %username% with the correct value
	//  depending on the single user login option
	ExpandEnvStrings(list,buffer,bufsize);

	// Get the first item...
	current = _tcstok(buffer,TEXT(";"));
	// Until there are no more items
	while(current != NULL)
	{
		// Current now has "DRIVE:\\Share\"
		if(path)
			free(path);

		// Path should be the first \ in the string
		// This was doing something very nasty (static +2) :)
		path = _tcsdup(wcsstr(current,TEXT("\\")));
		//_tcsncpy(drive,current,2);
		drive[0] = current[0];
		drive[1] = TEXT(':');
		drive[2] = TEXT('\0');

		EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Trying to Disconnect %s:%s"),drive,path);

		// All variables are there now - DO THE MAP
		if(!ImpersonateLoggedOnUser(hUser))
		{
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Impersonation during mapping failed."));
		}

		//EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Trying to map %s:%s with username: %s password: %s"),drive,path,settingsInfo->Username,settingsInfo->Password);
		lResult = WNetCancelConnection2(drive,CONNECT_UPDATE_PROFILE,TRUE);

        // Get next token
		current = _tcstok(NULL ,TEXT(";"));
	}

	// If we get this far, then all went well

	if(list)
		free(list);
	if(buffer)
		free(buffer);
	RevertToSelf();	// Reverse the impersonation
}

/**
	@brief	Save settings information to the registry

	The SaveSettings() method is called when information within the pGinaInfo structure 
	passed needs to be saved to the registry.  The information will be stored in the Windows 
	registry under the key \\HKLM\SOFTWARE\pGina\ and will be readable by all users and 
	writable only to local administrator accounts.

    @retval	boolean
	In the event of any errors, this method will provide all necessary warnings and 
	information to the user prior to returning a false value, otherwise a value of true
	will be returned.
*/
void SaveSettings(void)
{
    regWriteBool(saveLastUsername,_T("saveLastUsername"));
    regWriteBool(allowLockedLogout,_T("allowLockedLogout"));
	regWriteBool(encryptSettings,_T("encryptSettings"));
	regWriteDWORD(acctExpireTime,_T("acctExpireTime"));
	regWriteBool(acctCaching,_T("acctCaching"));
	regWriteString(windowTitle,_T("title"));
	regWriteString(motdText,_T("motd"));
	regWriteString(customLogo,_T("logo"));
	regWriteString(customLocked,_T("locked_logo"));
	regWriteString(pluginName,_T("pluginName"));
	regWriteString(UserLoginScript,_T("UserLoginScript"));
	regWriteString(sUsername,_T("sUsername"));  
	regWriteString(sPassword,_T("sPassword"));
	regWriteString(adminUsername,_T("adminUsername"));
	regWriteString(adminPassword,_T("adminPassword"));
	regWriteBool(manageDomain,_T("manageDomain"));
	regWriteBool(includeDomain,_T("includeDomain"));
	regWriteBool(scramblePasswords,_T("scramblePasswords"));
	regWriteBool(allowSpaces,_T("allowSpaces"));
	regWriteBool(disallowExpiration,_T("disallowExpiration"));
	regWriteBool(showUsernameAndAuth,_T("showUsernameAndAuth"));
	regWriteBool(showMachineName,_T("showMachineName"));
	regWriteBool(enableTask,_T("enableTask"));
	regWriteBool(enableLogout,_T("enableLogout"));
	regWriteBool(enableRestart,_T("enableRestart"));
	regWriteBool(disableAllEventLogs,_T("disableAllEventLogs"));
	regWriteBool(trueUnlock,_T("trueUnlock"));
	regWriteBool(honorLockedAccounts,_T("honorLockedAccounts"));
	regWriteBool(hookWlx,_T("hookWlx"));
	regWriteBool(honorAutoLogin,_T("honorAutoLogin"));
	regWriteString(settingsInfo.pathPlugin,_T("pathPlugin"));  
	regWriteString(settingsInfo.pathProfile,_T("pathProfile"));
	regWriteString(settingsInfo.pathProfileDefault, _T("pathProfileDefault"));
	regWriteString(settingsInfo.pathMSGina,_T("pathMSGina"));
	regWriteString(settingsInfo.mapPaths,_T("mapPaths"));
	regWriteString(defaultDomain,_T("defaultDomain"));
	regWriteString(globalGroups,_T("globalGroups"));
	regWriteBool(debugOutput,_T("debugOutput"));
	regWriteBool(keepProfiles,_T("keepProfiles"));
	regWriteBool(allowLocking,_T("allowLocking"));
	regWriteBool(mapAdmin,_T("mapAdmin"));
	regWriteBool(forceLocal,_T("forceLogin"));
	regWriteBool(hidePlugin,_T("hidePlugin"));
	regWriteBool(hideMotd,_T("hideMotd"));
	regWriteBool(displayErrors,_T("displayErrors"));
	regWriteBool(settingsInfo.disabled,_T("passThru"));
	regWriteBool(sUserLogin,_T("sUserLogin"));
	regWriteBool(disablePassword,_T("disablePassword"));
	regWriteString(sDomain,_T("sUserDomain"));
	regWriteDWORD(buttonAction,_T("buttonAction"));
	regWriteDWORD(defaultSessionTimeout,_T("defaultSessionTimeout"));
	regWriteBool(useBypass,_T("useBypass"));
    regWriteBool(idleLogoff,_T("idleLogoff"));
    regWriteDWORD(idleTime,_T("idleTime"));
    regWriteBool(idleAllowScreen,_T("idleAllowScreen"));
    regWriteBool(idleAllowLock,_T("idleAllowLock"));
    regWriteBool(idleAllowReturn,_T("idleAllowReturn"));
    regWriteBool(idleFromLocked,_T("idleFromLocked"));
	regWriteBool(pluginAutoLogin,_T("pluginAutoLogin"));
	regWriteString(palUsername,_T("palUsername"));       
	regWriteString(palPassword,_T("palPassword")); 
	regWriteBool(palShowError,_T("palShowError"));
	regWriteBool(quickLogin,_T("quickLogin"));
	regWriteString(quickUser,_T("quickUser"));
	regWriteString(quickPass,_T("quickPass"));
	regWriteString(quickCaption,_T("quickCaption"));
	regWriteBool(reqDom,_T("reqDom"));
	regWriteBool(keepAccount,_T("keepAccount"));
	regWriteBool(roamingProfile,_T("roamingProfile"));
} 
  
/**
	@brief	Load settings information from the registry

    The LoadSettings() method is called when settings need
    to be loaded from the registry.  The information will be stored in the Windows 
	registry under the key \\HKLM\Software\pGina\ and will be readable by all users and 
	writable only to local administrator accounts.

*/
void LoadSettings(void)
{
	// Clear mem first
	if(settingsInfo.pathProfile)
	{
		free(settingsInfo.pathProfile);
		settingsInfo.pathProfile = NULL;
	}
	if (settingsInfo.pathProfileDefault) {
		free(settingsInfo.pathProfileDefault);
		settingsInfo.pathProfileDefault = NULL;
	}
	if(settingsInfo.pathMSGina)
	{
		free(settingsInfo.pathMSGina);
		settingsInfo.pathMSGina = NULL;
	}
	if(settingsInfo.pathPlugin)
	{
		free(settingsInfo.pathPlugin);
		settingsInfo.pathPlugin = NULL;
	}
	if(defaultDomain)
	{
		free(defaultDomain);
		defaultDomain = NULL;
	}
	if(globalGroups)
	{
		free(globalGroups);
		globalGroups = NULL;
	}
	if(UserLoginScript)
	{
		free(UserLoginScript);
		UserLoginScript = NULL;
	}
	if(palUsername)
	{                          
			free(palUsername);
			palUsername = NULL;
	}                          
	if(palPassword)            
	{                          
			free(palPassword);
			palPassword = NULL;
	}  
	if(quickUser) {
		free(quickUser);
		quickUser = NULL;
	}
	if(quickPass) {
		free(quickPass);
		quickPass = NULL;
	}
	if(quickCaption) {
		free(quickCaption);
		quickPass = NULL;
	}

    saveLastUsername = false;
	sUserLogin = false;
	scramblePasswords = false;
	allowSpaces = true;
	disallowExpiration = false;
	manageDomain = false;
	includeDomain = false;
	trueUnlock = false;
	hookWlx = false;
	honorLockedAccounts = false;
	honorAutoLogin = false;
	usingAutoLogin = false;
	defaultSessionTimeout = 0;
	useBypass = false;
	acctCaching = false;
	acctExpireTime = 0;
	pluginAutoLogin = false;
	palShowError = false;
	quickLogin = false;
	reqDom = false;
	keepAccount = true;
	roamingProfile = false;
    allowLockedLogout = false;

    saveLastUsername = regReadBool(_T("saveLastUsername"));
    lastUserName = regReadString(_T("lastUserName"));
	reqDom = regReadBool(_T("reqDom"));
    allowLockedLogout = regReadBool(_T("allowLockedLogout"));
    quickLogin = regReadBool(_T("quickLogin"));
	quickUser = regReadString(_T("quickUser"));
	quickPass = regReadString(_T("quickPass"));
	quickCaption = regReadString(_T("quickCaption"));

	palShowError = regReadBool(_T("palShowError"));
	encryptSettings = regReadBool(_T("encryptSettings")); 	// THIS MUST ALWAYS BE FIRST
	acctExpireTime = regReadDWORD(_T("acctExpireTime"));
    idleTime = regReadDWORD(_T("idleTime"));
	acctCaching = regReadBool(_T("acctCaching"));

	settingsInfo.pathPlugin = regReadString(_T("pathPlugin"));
    pluginName = regReadString(_T("pluginName"));
    settingsInfo.pathProfile = regReadString(_T("pathProfile")); // Can be blank
	settingsInfo.pathProfileDefault = regReadString(_T("pathProfileDefault")); // Can be blank
	settingsInfo.pathMSGina = regReadString(_T("pathMSGina"));
	settingsInfo.mapPaths = regReadString(_T("mapPaths"));		 // Can be blank
	defaultDomain = regReadString(_T("defaultDomain"));			 // Can be blank
	globalGroups = regReadString(_T("globalGroups"));			 // Can be blank
	customLocked = regReadString(_T("locked_logo"));
	customLogo = regReadString(_T("logo"));
    adminUsername = regReadString(_T("adminUsername"));
	adminPassword = regReadString(_T("adminPassword"));
    debugOutput = regReadBool(_T("debugOutput"));
	keepProfiles = regReadBool(_T("keepProfiles"));
	settingsInfo.disabled = regReadBool(_T("passThru"));
	allowLocking = regReadBool(_T("allowLocking"));
	mapAdmin = regReadBool(_T("mapAdmin"));
	forceLocal = regReadBool(_T("forceLogin"));                                         
	displayErrors = regReadBool(_T("displayErrors"));                                                   
	honorLockedAccounts = regReadBool(_T("honorLockedAccounts"));
	hideMotd = regReadBool(_T("hideMotd"));                                                 
	hidePlugin = regReadBool(_T("hidePlugin"));
	sUserLogin = regReadBool(_T("sUserLogin"));
	sUsername = regReadString(_T("sUsername"));
	sPassword = regReadString(_T("sPassword"));
	UserLoginScript = regReadString(_T("UserLoginScript"));
    scramblePasswords = regReadBool(_T("scramblePasswords"));
	allowSpaces = regReadBool(_T("allowSpaces"));
	disallowExpiration = regReadBool(_T("disallowExpiration"));
    manageDomain = regReadBool(_T("manageDomain"));
	includeDomain = regReadBool(_T("includeDomain"));
	disableAllEventLogs = regReadBool(_T("disableAllEventLogs"));
	buttonAction = regReadDWORD(_T("buttonAction"));
    showUsernameAndAuth = regReadBool(_T("showUsernameAndAuth"));
	showMachineName = regReadBool(_T("showMachineName"));
	enableTask = regReadBool(_T("enableTask"));
	enableLogout = regReadBool(_T("enableLogout"));
	enableRestart = regReadBool(_T("enableRestart"));
	disablePassword = regReadBool(_T("disablePassword"));
	trueUnlock = regReadBool(_T("trueUnlock"));
    honorAutoLogin = regReadBool(_T("honorAutoLogin"));
    sDomain = regReadString(_T("sUserDomain"));
    hookWlx = regReadBool(_T("hookWlx"));
    defaultSessionTimeout = regReadDWORD(_T("defaultSessionTimeout"));
    useBypass = regReadBool(_T("useBypass"));
    idleLogoff = regReadBool(_T("idleLogoff"));
    idleAllowScreen = regReadBool(_T("idleAllowScreen"));
    idleAllowLock = regReadBool(_T("idleAllowLock"));
    idleAllowReturn = regReadBool(_T("idleAllowReturn"));
    idleFromLocked = regReadBool(_T("idleFromLocked"));
	pluginAutoLogin = regReadBool(_T("pluginAutoLogin"));
	palUsername = regReadString(_T("palUsername"));       
	palPassword = regReadString(_T("palPassword")); 
	keepAccount = regReadBool(_T("keepAccount"));
	roamingProfile = regReadBool(_T("roamingProfile"));

	if(keepProfiles) {
		keepAccount = true;	// If Keep Profiles is turned on, this MUST be
	}
}

/**
	@brief	Writes a string to the registry
	Writes the passed string to the indicated value under HKLM\\SOFTWARE\\pGina
	
	@param  value       String to store
	@param  location    Name of the value to save
*/
void regWriteString(LPTSTR value,LPCTSTR location)
{
	HKEY hKey;

	if(encryptSettings) value = encryptString(value);

	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,_T("SOFTWARE\\pGina"),0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,&hKey,NULL) == ERROR_SUCCESS)
	{
		if(value != NULL)
		{
			if(RegSetValueEx(hKey,location,0,REG_SZ,(CONST BYTE *) value, _tcslen(value)*sizeof(TCHAR)) == ERROR_SUCCESS)
			{
				EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Wrote registry entry %s with value %s"),location,value);
				RegFlushKey(hKey);
				RegCloseKey(hKey);
                return;
			}
		}
		else
		{
			if(RegSetValueEx(hKey,location,0,REG_SZ,(CONST BYTE *) _T("\0"), _tcslen(_T("\0"))*sizeof(TCHAR)) == ERROR_SUCCESS)
			{
				EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Wrote registry entry %s with value %s"),location,value);
				RegFlushKey(hKey);
				RegCloseKey(hKey);
				return;
			}
		}
	}

	EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Writing to registry entry %s with value %s failed"),location,value);
	RegFlushKey(hKey);
	RegCloseKey(hKey);
}

/**
	@brief	Writes a boolean (as dword) to the registry
	@author Nathan Yocom
	
	Writes the passed boolean to the indicated value under HKLM\\SOFTWARE\\pGina
	
	@param  value       Boolean to store
	@param  location    Name of the value to save
	
*/
void regWriteBool(BOOL value,LPCTSTR location)
{
	HKEY hKey;
	DWORD write;

	write = (value ? 1 : 0);

	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,_T("SOFTWARE\\pGina"),0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,  &hKey,NULL) == ERROR_SUCCESS)
		if(RegSetValueEx(hKey,location,0,REG_DWORD,(const BYTE*)&write,sizeof(DWORD)) == ERROR_SUCCESS)
		{
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Wrote registry entry %s with value %d"),location,write);
			RegFlushKey(hKey);
			RegCloseKey(hKey);
			return;
		}
	EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Writing to registry entry %s with value %d failed"),location,write);
	RegFlushKey(hKey);
	RegCloseKey(hKey);
	return;
}

/**
	@brief	Writes a dword to the registry
	@author Nathan Yocom
	
	Writes the passed DWORD to the indicated value under HKLM\\SOFTWARE\\pGina
	
	@param  value       dword to store
	@param  location    Name of the value to save
*/
void regWriteDWORD(DWORD value,LPCTSTR location)
{
	HKEY hKey;
	DWORD write;

	write = value;

	if(RegCreateKeyEx(HKEY_LOCAL_MACHINE,_T("SOFTWARE\\pGina"),0,NULL,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,NULL,  &hKey,NULL) == ERROR_SUCCESS)
		if(RegSetValueEx(hKey,location,0,REG_DWORD,(const BYTE*)&write,sizeof(DWORD)) == ERROR_SUCCESS)
		{
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Wrote registry entry %s with value %d"),location,write);
			RegFlushKey(hKey);
			RegCloseKey(hKey);
			return;
		}
	EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Writing to registry entry %s with value %d failed"),location,write);
	RegFlushKey(hKey);
	RegCloseKey(hKey);
	return;
}

/**
	@brief	Reads a string to the registry
	@author Nathan Yocom
	
	Reads the indicated location under HKLM\\Software\\pGina and return it
	
	@param  location    Name of the value to read
	
	@retval TCHAR*  A newly allocated TCHAR[] with the read value (up to 6000 chars) or NULL
*/
TCHAR * regReadString(LPCTSTR location)
{
	LONG lResult;
	HKEY hKey;
	DWORD dwBytesRead = 6000;
	TCHAR result[6000];
	TCHAR *ret;

	lResult = RegOpenKeyEx(HKEY_LOCAL_MACHINE,_T("SOFTWARE\\pGina"),0,KEY_QUERY_VALUE,&hKey);

	if(lResult == ERROR_SUCCESS)
	{
		lResult = RegQueryValueEx(hKey,location,NULL,NULL,(LPBYTE)&result,&dwBytesRead);

		if(lResult == ERROR_SUCCESS)
		{
			if(dwBytesRead > 0)
			{
				EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Read registry entry %s result: *%s*, size: %d"),location,result,dwBytesRead);

				if(!_tcscmp(result,TEXT(""))) // If its just a space
				{
					EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Read registry entry %s result: NULL"),location);                
					RegCloseKey(hKey);
					return NULL;                                                                                 
				}
				else
				{
					if(encryptSettings) {
						ret = decryptString(result);
					} else {
						ret = _tcsdup(result);
					}
					RegCloseKey(hKey);
					return ret;
				}
			}
			else
			{
				EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Read registry entry %s result: NULL"),location);
				RegCloseKey(hKey);
				return NULL;
			}
		}
		else
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Error reading registry entry %s."),location);
	}
	RegCloseKey(hKey);
	return NULL;
}

/**
	@brief	Load a bool value from the HKLM/pGina registry
	@author Nathan Yocom
	
	@param location The key to load
    @retval boolean value, default to false
*/
BOOL regReadBool(LPCTSTR location)
{
	LONG lResult;
	HKEY hKey;
	DWORD value = 0;
	DWORD size = sizeof(DWORD);

	lResult = RegOpenKeyEx(HKEY_LOCAL_MACHINE,_T("SOFTWARE\\pGina"),0,KEY_QUERY_VALUE,&hKey);

	if(lResult == ERROR_SUCCESS)
	{
		lResult = RegQueryValueEx(hKey,location,NULL,NULL,(LPBYTE)&value,&size);

		if(lResult == ERROR_SUCCESS)
		{
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Read registry entry %s result: %d"),location,value);
			RegCloseKey(hKey);
			if(value == 1)
				return true;
			else
				return false;
		}
		else
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Error reading registry entry %s."),location);
	}
	RegCloseKey(hKey);
	return false;
}

/**
	@brief	Load a dword value from the HKLM/pGina registry
	@author Nathan Yocom
	
	@param location The key to load
    @reval dword read, default to 0 
*/
DWORD regReadDWORD(LPCTSTR location)
{
	LONG lResult;
	HKEY hKey;
	DWORD value = 0;
	DWORD size = sizeof(DWORD);

	lResult = RegOpenKeyEx(HKEY_LOCAL_MACHINE,_T("SOFTWARE\\pGina"),0,KEY_QUERY_VALUE,&hKey);

	if(lResult == ERROR_SUCCESS)
	{
		lResult = RegQueryValueEx(hKey,location,NULL,NULL,(LPBYTE)&value,&size);

		if(lResult == ERROR_SUCCESS)
		{
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Read registry entry %s result: %d"),location,value);
			RegCloseKey(hKey);
			return value;
		}
		else
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Error reading registry entry %s."),location);
	}
	RegCloseKey(hKey);
	return 0;
}


/**
	@brief	Writes passed info to pGina Event Log
	@author	Nathan Yocom
	
	@param  debug   A boolean indicating that the info is to be printed only if debugOutput is on (true), or always (false)
	@param  log_type    The EVENT type for ReportEvent
	@param	string	The string (TCHAR []) to be output to the file
*/
void EventLog(BOOL debug,WORD log_type,TCHAR *string,...)
{
	LPCWSTR *strings[1];
	LPCWSTR buffer[30000];
	int number = 0;
	va_list argList;
	HANDLE myEventSource;

	// Don't do anything if logs are totally disabled
	if(disableAllEventLogs)
		return;

    if(windowsVersion != WINDOWS_NT_4 && ((debug && debugOutput) || !debug))
	{
		va_start(argList,string);
		vswprintf((wchar_t *)buffer,(const wchar_t *)string,argList);
		strings[0] = (LPCWSTR *)&buffer;

		
		if(log_type == EVENTLOG_AUDIT_SUCCESS || log_type == EVENTLOG_AUDIT_FAILURE)
			myEventSource = RegisterEventSource(NULL,TEXT("Security"));
		else
			myEventSource = RegisterEventSource(NULL,TEXT("pGina"));

   		ReportEvent(myEventSource,log_type,0,0,NULL,1,0,(LPCWSTR *)strings,NULL);
   		DeregisterEventSource(myEventSource);
	} // Logging is not available in NT 4
}

/**
	@brief	Sets the credentials for AutoLogin
	@author	Jeremy C. Sparks (Willamette University), jcsparks@willamette.edu
	@note	Code re-used with permission from WuGina project, by Jeremy C. Sparks.
	@version	1.7.3
	
	Sets the username and password portions of the autologin registry.
	
	@param 	AutoUser	    The username to autologin with
	@param	AutoPassword	The password to use
*/
void SetAutoCredentials(PTCHAR AutoUser, PTCHAR AutoPassword)
{
	HKEY    hKey;
	LONG    lResult;
	PTCHAR  pszSystemName;
	TCHAR result[2048];
	DWORD dwBytesRead = 2048;
	LPTSTR domainToUse;

	pszSystemName = GlobalAlloc( GPTR, 2048 );
	if( ! pszSystemName )
		return;

	GetEnvironmentVariable( _T("COMPUTERNAME"), pszSystemName, 2048 );

	lResult = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
							_T("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon"),
							0, KEY_QUERY_VALUE, &hKey );

	if(lResult == ERROR_SUCCESS)
	{
		// Check to see if a value is there already
		lResult = RegQueryValueEx(hKey,_T("DefaultDomainName"),NULL,NULL,(LPBYTE)&result,&dwBytesRead);
		if(lResult != ERROR_SUCCESS)
		{
			dwBytesRead = 0;
		}

		RegCloseKey( hKey );
	}

	lResult = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
							_T("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon"),
							0, KEY_WRITE, &hKey );

	if(lResult == ERROR_SUCCESS)
	{
		lResult = RegSetValueEx( hKey, _T("DefaultUserName"), 0, REG_SZ,
								 (CONST BYTE*)AutoUser,
								 _tcslen( AutoUser ) * sizeof(_TCHAR)
							   );
		if(lResult == ERROR_SUCCESS)
		{
			lResult = RegSetValueEx( hKey, _T("DefaultPassword"), 0, REG_SZ,
									 (CONST BYTE*)AutoPassword,
									 _tcslen( AutoPassword ) * sizeof(_TCHAR)
								   );
			if(lResult == ERROR_SUCCESS)
			{
				// What domain name to use...
				if(settingsInfo.authType == AUTH_LOCAL)
					domainToUse = pszSystemName; // Local user
				else
				{
					// If it is a domain user, then use the proper domain
					if(settingsInfo.authType == AUTH_DOMAIN)
					{
						if(settingsInfo.defaultDomain)
							domainToUse = settingsInfo.defaultDomain;
						else
							domainToUse	= defaultDomain;
					}
					else // Plugin authenticated
					{
						// Was the account created on the domain
						if(manageDomain || includeDomain)
						{
							// use a domain
							if(settingsInfo.defaultDomain)
								domainToUse = settingsInfo.defaultDomain;
							else
								domainToUse	= defaultDomain;    
						}
						else // Local machine was used
							domainToUse	= pszSystemName;
					}
				}

				// Allow single user login to be done via a Domain, 
				if(sUserLogin && sDomain && settingsInfo.authType == AUTH_PLUGIN)
				{
					domainToUse = sDomain;
				}

				lResult = RegSetValueEx( hKey, _T("DefaultDomainName"), 0, REG_SZ,
										 (CONST BYTE*)domainToUse,
										 _tcslen( domainToUse ) * sizeof(_TCHAR)
									   );

				if(lResult == ERROR_SUCCESS)
				{
					lResult= RegFlushKey( hKey );
				}
				else
					EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Error writing AutoAdmin information to registry."));
			}
		}

		RegFlushKey( hKey );
		RegCloseKey( hKey );
	}
	if( pszSystemName )	GlobalFree( pszSystemName );

}

/**
	@brief	Sets or removes the AutoAdmin bit
	@author	Jeremy C. Sparks (Willamette University), jcsparks@willamette.edu
	@note	Code re-used with permission from WuGina project, by Jeremy C. Sparks.
	@version	1.7.3
	
	Sets or removes the autologin registry bits.
	
	@param 	value	1 to set AutoAdmin true (use SetAutoCredentials() to set the username/password)
	0 to remove AutoAdmin bits and remove AutoCredentials (if they exist)
*/
void SetAutoAdmin(int value)
{ 
	HKEY    hKey;
	LONG    lResult;
	TCHAR result[2048];
	DWORD dwBytesRead = 2048;
	TCHAR *res;

	lResult = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
							_T("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon"),
							0, KEY_WRITE, &hKey );

	if(lResult == ERROR_SUCCESS)
	{
		lResult = RegSetValueEx( hKey, _T("AutoAdminLogon"), 0, REG_SZ,
								 value ? "1" : "0",
								 1 * sizeof(TCHAR));
		lResult = RegSetValueEx( hKey, _T("ForceAutoLogon"), 0, REG_SZ, "1", 1 * sizeof(TCHAR));

		// clear any leftover entries.
		// RegDeleteValue( hKey, _T("DefaultUserName") ); // We leave this now, to fix an issue with Ghost Console, and to allow us to use it for displaying last logged in user
		RegDeleteValue( hKey, _T("DefaultPassword") );

		if(lResult == ERROR_SUCCESS)
		{
			lResult= RegFlushKey( hKey );
		}
		else
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Error writing AutoAdmin bit to registry."));

		RegCloseKey( hKey );
	}

	// If we are setting stuff up, check to see if DefaultDomainName exists
	//  if it does, save it to DefaultDomainNameSAVE and restore it later
	if(value)
	{
		lResult = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
								_T("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon"),
								0, KEY_QUERY_VALUE, &hKey);

		if(lResult == ERROR_SUCCESS)
		{
			// Check to see if a value is there already
			lResult = RegQueryValueEx(hKey,_T("DefaultDomainName"),NULL,NULL,(LPBYTE)&result,&dwBytesRead);

			if(lResult == ERROR_SUCCESS && dwBytesRead > 0)
			{
				// save the string
				regWriteString(result,_T("DefaultDomainNameSAVE"));
			}

			RegCloseKey( hKey );
		}
	}
	else
	{
		res = regReadString(_T("DefaultDomainNameSAVE"));

		if(res != NULL)	// restore the string
			RegSetValueEx(hKey,_T("DefaultDomainName"),0,REG_SZ,(CONST BYTE *) res, _tcslen(res)*sizeof(TCHAR));
		else // or restore NULL by deleting the key
			RegDeleteValue( hKey, _T("DefaultDomainName"));
	}
}

/**
	@brief	Creates a user account on the local machine
	@note	Account creation fails if user already exists

	Creates a local user account with the username and password in the pGinaInfo structure
	settingsInfo.
*/
BOOL CreateLocalAccount(void)
{
	USER_INFO_1 ui;
	DWORD dwLevel = 1;
	DWORD dwError = 0;
	NET_API_STATUS nStatus;
	USER_INFO_1003 new_passwd;   
	USER_INFO_1007 new_comment;   
	USER_INFO_1011 new_fullname; 
	USER_INFO_1017 new_acctExpires;
	LOCALGROUP_MEMBERS_INFO_3 memberInfo;
	LPTSTR tempFullname;
	LPTSTR tempDescription;
	LPTSTR tempUsername;
	LPTSTR tempPassword;
	LPTSTR domainToUse;
	LPTSTR serverToUse = NULL;
	HANDLE hUser;

	SID_IDENTIFIER_AUTHORITY siaNT = SECURITY_NT_AUTHORITY;
	PSID pSidAdmins = NULL;
	WCHAR szGroupName[UNLEN + 1];
	WCHAR szDomainName[DNLEN + 1];
	ULONG cchGroupName = UNLEN + 1;
	ULONG cchDomainName = DNLEN + 1;
	SID_NAME_USE Use;

	tempUsername = _tcsdup(settingsInfo.Username);
	tempPassword = _tcsdup(settingsInfo.Password);
	tempFullname = _tcsdup(settingsInfo.userFullName);
	tempDescription = _tcsdup(settingsInfo.userDescription);


	ui.usri1_name = (LPWSTR) tempUsername;
	memberInfo.lgrmi3_domainandname = (LPWSTR) tempUsername;
	ui.usri1_password = (LPWSTR) tempPassword;
	if(disallowExpiration)
		ui.usri1_flags = UF_SCRIPT | UF_PASSWD_CANT_CHANGE | UF_DONT_EXPIRE_PASSWD;
	else
		ui.usri1_flags = UF_SCRIPT | UF_PASSWD_CANT_CHANGE;

	ui.usri1_script_path = NULL;    
	ui.usri1_priv = USER_PRIV_USER;
	ui.usri1_comment = NULL;
	ui.usri1_home_dir = NULL;
	ui.usri1_password_age = 0;

	// Determine what domain to do all this against
	if(!manageDomain)
		domainToUse = TEXT(".");	// Local machine
	else if(manageDomain)
	{
		if(settingsInfo.defaultDomain)
			domainToUse = settingsInfo.defaultDomain;
		else
			domainToUse	= defaultDomain;
	}

	// If we are to manage the domain, we must first login as a domain admin
	if(manageDomain)
	{
		// make sure we can login as the domain admin
		if(!LogonUser(adminUsername,domainToUse,adminPassword,LOGON32_LOGON_INTERACTIVE,LOGON32_PROVIDER_DEFAULT,&hUser))
		{
			// Login failed
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Logging in as an admin to domain %s with username %s failed."),domainToUse,adminUsername);
			return false;
		}
		else
		{
			// Otherwise we are in, Impersonate this user...
			ImpersonateLoggedOnUser(hUser);
			// And setup the correct machine name for our Net functions
			serverToUse = getDNSfromDomain();
		}
	}

	EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Adding account with username: %s"),ui.usri1_name);

	// Try and add the account
	nStatus = NetUserAdd(serverToUse,1,(LPBYTE)&ui,NULL);

	if(nStatus == NERR_Success || nStatus == NERR_UserExists) // Checking for Exists here allows us to set info, groups etc on existing users
	{
		if(nStatus == NERR_UserExists)
		{
			if(forceLocal)
			{
				EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Forcing login for %s"),settingsInfo.Username);

				new_passwd.usri1003_password = tempPassword;

				nStatus = NetUserSetInfo(serverToUse,(LPCWSTR) tempUsername,1003,(LPBYTE) &new_passwd,NULL);

				if(nStatus != NERR_Success)
					EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Error forcing login, still need to work on recovery from: nStatus = %d"),nStatus);
				else // Their password is set, work on adding them to the app. groups
					EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("User account forced - password reset accordingly."));
			}
		}

		if(settingsInfo.isAdmin)
		{
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Setting user %s as an admin"),ui.usri1_name);

			// Get the SID of the Admin group
			AllocateAndInitializeSid(&siaNT, 2, SECURITY_BUILTIN_DOMAIN_RID,DOMAIN_ALIAS_RID_ADMINS, 0, 0, 0, 0, 0, 0, &pSidAdmins);
			// Use that SID to get the group name
			LookupAccountSidW(NULL, pSidAdmins, szGroupName, &cchGroupName,szDomainName, &cchDomainName, &Use);

			nStatus = NetLocalGroupAddMembers(serverToUse,szGroupName,3,(LPBYTE) &memberInfo,1);

			if(nStatus != NERR_Success)
				EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Failed to set %s as an admin"),ui.usri1_name);
		}

		if(tempFullname != NULL)
		{
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Setting users fullname: %s"),tempFullname);

			new_fullname.usri1011_full_name = (LPWSTR) tempFullname;

			nStatus = NetUserSetInfo(serverToUse,(LPWSTR)tempUsername,1011,(LPBYTE) &new_fullname,NULL);

			if(nStatus != NERR_Success)
				EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Settings users fullname failed."));
		}
		else
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("No Fullname for user"));

		if(tempDescription != NULL)
		{
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Setting users description: %s"),tempDescription);
			new_comment.usri1007_comment = (LPWSTR) tempDescription;

			nStatus = NetUserSetInfo(serverToUse,(LPWSTR) tempUsername,1007,(LPBYTE) &new_comment, NULL);

			if(nStatus != NERR_Success)
				EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Failed to set users description."));
		}
		else
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("No description for user."));

		if (acctCaching && acctExpireTime!=0 ){
			LPTIME_OF_DAY_INFO pBuf = NULL;
			DWORD timeToUse;
			timeToUse = acctExpireTime;
			nStatus = NetRemoteTOD(serverToUse,(LPBYTE *)&pBuf);
			if (nStatus == NERR_Success)
   			{
      			if (pBuf != NULL)
      			{
      				new_acctExpires.usri1017_acct_expires = pBuf->tod_elapsedt + timeToUse;
      				NetApiBufferFree(pBuf);

					EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Setting account expiration: %d"),timeToUse);
					nStatus = NetUserSetInfo(serverToUse,(LPWSTR) tempUsername,1017,(LPBYTE) &new_acctExpires, NULL);
					if(nStatus != NERR_Success)
						EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Failed to set users account expiration."));    				
         		}
         	}else{
         		EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Failed to get TimeOfDay."));
        	}
		}

		if(settingsInfo.userGroups != NULL || globalGroups != NULL)
		{
			AddUserToGroups();
		}
	}
	else
	{
		EventLog(true,EVENTLOG_ERROR_TYPE,TEXT("Adding user %s failed with unhandled error, nStatus = %d, dwError = %d"),ui.usri1_name,nStatus,dwError);

		if(manageDomain)
		{
			RevertToSelf();
			CloseHandle(hUser);
			if(serverToUse)
				free(serverToUse);
		}
		return false; // Return error
	} 

	if(manageDomain)
	{
		RevertToSelf();
		CloseHandle(hUser);
		if(serverToUse)
			free(serverToUse);
	}

	free(tempUsername);
	free(tempPassword);
	free(tempFullname);
	free(tempDescription);
	serverToUse = NULL;
	return true;
}

/**
	@brief	Deletes a user account on the local machine
	@author	Nathan Yocom

	Deletes a local user account with the username in the pGinaInfo structure
	settingsInfo.
	
*/
void DeleteLocalAccount(BOOL deleteUser)
{
	NET_API_STATUS nStatus = NERR_Success;
	TCHAR pathProfile[2048];
	LPTSTR tempUsername;
	DWORD dwLevel = 4;
	PTCHAR  userProfile;
	LPTSTR domainToUse;
	LPTSTR serverToUse = NULL;
	HANDLE hUser;

	EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("in delete"));

	// Determine what domain to do all this against
	if(!manageDomain)
		domainToUse = TEXT(".");	// Local machine
	else if(manageDomain)
	{
		if(settingsInfo.defaultDomain)
			domainToUse = settingsInfo.defaultDomain;
		else
			domainToUse	= defaultDomain;
	}

	// If we are to manage the domain, we must first login as a domain admin
	if(manageDomain)
	{
		// make sure we can login as the domain admin
		if(!LogonUser(adminUsername,domainToUse,adminPassword,LOGON32_LOGON_INTERACTIVE,LOGON32_PROVIDER_DEFAULT,&hUser))
		{
			// Login failed
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Logging in as an admin to domain %s with username %s failed."),domainToUse,adminUsername);
			return;
		}
		else
		{
			// Otherwise we are in, Impersonate this user...
			ImpersonateLoggedOnUser(hUser);
			serverToUse = getDNSfromDomain();
		}
	}

	tempUsername = _tcsdup(settingsInfo.Username);
	userProfile = GlobalAlloc( GPTR, 2048 );

	if(userProfile)
		GetEnvironmentVariable( _T("USERPROFILE"), userProfile, 2048 );

	EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("user profile at: %s"),userProfile);

	// If we should remove the user account, do so
	if(deleteUser) {
		nStatus = NetUserDel(serverToUse,tempUsername);
		if(nStatus == NERR_Success) {
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Deletion of account for %s was successfull"),settingsInfo.Username);
		} else { 
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Deletion of account for %s failed."),settingsInfo.Username);
		}
	}

	// Remove the users profile
    if(userProfile) {
		EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Deleting directory: %s"),userProfile);
		DeleteFolder(userProfile);
	} else { // Default to C:\Documents and Settings 
		EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Deleting directory: C:\\Documents and Settings\\%s"),settingsInfo.Username);
		_stprintf(pathProfile,TEXT("C:\\Documents and Settings\\%s"),tempUsername);
		DeleteFolder(pathProfile);
	}
    
	if(userProfile)
		GlobalFree(userProfile);
	free(tempUsername);

	if(manageDomain)
	{
		RevertToSelf();
		CloseHandle(hUser);
		if(serverToUse)
			free(serverToUse);
	}
}

/**
	@brief	Deletes a folder recursivley
	@author	Nathan Yocom
	
	Deletes the passed folder recursivley (everything including .? files)
	
	@param directory A wchar array to the directory to be deleted
*/
void DeleteFolder(LPCTSTR directory)
{
	WIN32_FIND_DATA FileInformation;
	HANDLE hFile;
	TCHAR fullPath[2048];

	_tcscpy(fullPath,directory);
	_tcscat(fullPath,TEXT("\\*.*"));

	hFile = FindFirstFile(fullPath,&FileInformation);
	if(hFile != INVALID_HANDLE_VALUE)
	{
		do
		{
			if((((FileInformation.cFileName[0]) == '.') &&
				((FileInformation.cFileName[1]) != '.') &&
				((FileInformation.cFileName[1]) != '\0'))
			   || ((FileInformation.cFileName[0]) != '.'))
			{
				_tcscpy(fullPath,directory);
				_tcscat(fullPath,TEXT("\\"));
				_tcscat(fullPath,FileInformation.cFileName);

				if(FileInformation.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				{
					DeleteFolder((LPCTSTR)fullPath);    
				}
				else
				{
					if(SetFileAttributes(fullPath,FILE_ATTRIBUTE_NORMAL) == FALSE)
						return;

					if(DeleteFile(fullPath) == FALSE)
						return;

				}
			}
		} while(FindNextFile(hFile,&FileInformation) == TRUE);

		FindClose(hFile);

		if(SetFileAttributes(directory,FILE_ATTRIBUTE_NORMAL) == FALSE)
			return;

		if(RemoveDirectory(directory) == FALSE)
			return;

	}
}

/**
	@brief	Dumps any and all state info to the given buffer
	@author	Nathan Yocom
	@todo should check the size of strings as we go to avoid overflow
	@todo Add a dump of the pGina event logs
*/
void dumpState(wchar_t *buffer, unsigned int size)
{
	wchar_t *temp = NULL;
	GetMotd();
	GetTitle();
	LoadLogo();
	temp = malloc(4096 * sizeof(wchar_t));
	memset(buffer,'\0',size);
	swprintf(buffer,_T("Windows Build: %d.%d %s (Build %d)\r\n"), osvi.dwMajorVersion, osvi.dwMinorVersion,osvi.szCSDVersion, osvi.dwBuildNumber &0xFFFF);
	swprintf(temp,_T("pGina Version: %s\r\n"),VERSION_SHORT);
	wcscat(buffer,temp);
	swprintf(temp,_T("============ pGina State =============\r\n"));
	wcscat(buffer,temp);
	swprintf(temp,_T("pGinahInstance: 0x%08x\r\n"), pGinahInstance);
	wcscat(buffer,temp);
	swprintf(temp,_T("hUser: 0x%08x\r\n"), hUser);
	wcscat(buffer,temp);
	swprintf(temp,_T("debugOutput: %s\r\n"), debugOutput ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("keepProfiles: %s\r\n"), keepProfiles ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("allowLocking: %s\r\n"), allowLocking ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("mapAdmin: %s\r\n"), mapAdmin ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("buttonAction: %d\r\n"), buttonAction);
	wcscat(buffer,temp);
	swprintf(temp,_T("hideMotd: %s\r\n"), hideMotd ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("hidePlugin: %s\r\n"), hidePlugin ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("displayErrors: %s\r\n"), displayErrors ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("isRequired: %s\r\n"), (pluginIsRequired ? pluginIsRequired() : false) ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("forceLocal: %s\r\n"), forceLocal ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("defaultDomain: %s\r\n"), defaultDomain);
	wcscat(buffer,temp);
	swprintf(temp,_T("windowTitle: [%s]\r\n"), windowTitle);
	wcscat(buffer,temp);
	swprintf(temp,_T("motdText: [%s]\r\n"), motdText);
	wcscat(buffer,temp);
	swprintf(temp,_T("customLogo: %s\r\n"), customLogo);
	wcscat(buffer,temp);
	swprintf(temp,_T("globalGroups: %s\r\n"), globalGroups);
	wcscat(buffer,temp);
	swprintf(temp,_T("sUsername: %s\r\n"), sUsername ? _T("Set") : _T("Not Set"));
	wcscat(buffer,temp);
	swprintf(temp,_T("sPassword: %s\r\n"), sPassword ? _T("Set") : _T("Not Set"));
	wcscat(buffer,temp);
	swprintf(temp,_T("sDomain: %s\r\n"), sDomain ? _T("Set") : _T("Not Set"));
	wcscat(buffer,temp);
	swprintf(temp,_T("sUserLogin: %s\r\n"), sUserLogin ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("scramblePasswords: %s\r\n"), scramblePasswords ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("allowSpaces: %s\r\n"), allowSpaces ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("UserLoginScript: %s\r\n"), UserLoginScript);
	wcscat(buffer,temp);
	swprintf(temp,_T("disallowExpiration: %s\r\n"), disallowExpiration ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("isLoggedIn: %s\r\n"), isLoggedIn ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("includeDomain: %s\r\n"), includeDomain ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("manageDomain: %s\r\n"), manageDomain ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("adminUsername: %s\r\n"), adminUsername ? _T("Set") : _T("Not Set"));
	wcscat(buffer,temp);
	swprintf(temp,_T("adminPassword: %s\r\n"), adminPassword ? _T("Set") : _T("Not Set"));
	wcscat(buffer,temp);
	swprintf(temp,_T("showUsernameAndAuth: %s\r\n"), showUsernameAndAuth ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("showMachineName: %s\r\n"), showMachineName ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("enableTask: %s\r\n"), enableTask ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("enableLogout: %s\r\n"), enableLogout ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("enableRestart: %s\r\n"), enableRestart ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("windowsVersion: %d\r\n"), windowsVersion);
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->pathMSGina: %s\r\n"), settingsInfo.pathMSGina);
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->pathPlugin: %s\r\n"), settingsInfo.pathPlugin);
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->mapPaths: %s\r\n"), settingsInfo.mapPaths);
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->Username: %s\r\n"), settingsInfo.Username ? _T("Set") : _T("Not Set"));
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->Password: %s\r\n"), settingsInfo.Password ? _T("Set") : _T("Not Set"));
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->isAdmin: %s\r\n"), settingsInfo.isAdmin ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->disabled: %s\r\n"), settingsInfo.disabled ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->authType: %d\r\n"), settingsInfo.authType);
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->userGroups: %s\r\n"), settingsInfo.userGroups);
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->userDescription: %s\r\n"), settingsInfo.userDescription);
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->userFullName: %s\r\n"), settingsInfo.userFullName ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->allowPassChange: %s\r\n"), settingsInfo.allowPassChange ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->errorString: %s\r\n"), settingsInfo.errorString);
	wcscat(buffer,temp);
	swprintf(temp,_T("SI->defaultDomain: %s\r\n"), settingsInfo.defaultDomain);
	wcscat(buffer,temp);
	swprintf(temp,_T("True Unlock: %s\r\n"), trueUnlock ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	swprintf(temp,_T("WLX Hook: %s\r\n"), hookWlx ? _T("True") : _T("False"));
	wcscat(buffer,temp);
	LoadPlugin();
	swprintf(temp,_T("AboutPlugin: %s\r\n"), pluginAboutPlugin());
	wcscat(buffer,temp);
}

/** 
		@brief  Adds user to a list of groups   
   
		Add the user to all of the groups listed in settingsInfo.userGroups.  The list   
		should be semi-colon seperated and all groups must exist.   
   
		@bug    _tcsdup's all over the place inside a loop... THIS NEEDS CLEANUP BAD!   
*/   
void AddUserToGroups(void)   
{
	LOCALGROUP_MEMBERS_INFO_3 memberInfo;   
	NET_API_STATUS nStatus;   
	LOCALGROUP_INFO_1         localgroup_info;

	LPTSTR currentGroup;   
	LPTSTR Username;
	LPTSTR domainToUse;
	LPTSTR serverToUse = NULL;
	HANDLE hUser;

	TCHAR *fullList = NULL;   
	TCHAR *location = NULL;   

	// Setup our username for conversion   
	Username = _tcsdup(settingsInfo.Username);   
	// Setup the username for later use   
	memberInfo.lgrmi3_domainandname = Username;   

    // Determine what domain to do all this against
	if(!manageDomain)
		domainToUse = TEXT(".");	// Local machine
	else if(manageDomain)
	{
		if(settingsInfo.defaultDomain)
			domainToUse = settingsInfo.defaultDomain;
		else
			domainToUse	= defaultDomain;
	}

	// If we are to manage the domain, we must first login as a domain admin
	if(manageDomain)
	{
		// make sure we can login as the domain admin
		if(!LogonUser(adminUsername,domainToUse,adminPassword,LOGON32_LOGON_INTERACTIVE,LOGON32_PROVIDER_DEFAULT,&hUser))
		{
			// Login failed
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Logging in as an admin to domain %s with username %s failed when attempting to add groups"),domainToUse,adminUsername);
			return;
		}
		else
		{
			// Otherwise we are in, Impersonate this user...
			ImpersonateLoggedOnUser(hUser);
			serverToUse = getDNSfromDomain();
		}
	}

    if(globalGroups != NULL)
	{
		fullList = _tcsdup(globalGroups);
		// Pull of the first in the list   
		location = _tcstok(fullList,TEXT(";"));

		while(location != NULL)
		{
			currentGroup = _tcsdup(location);   

			nStatus = NetLocalGroupAddMembers(serverToUse,currentGroup,3,(LPBYTE) &memberInfo,1);

			// Try adding the group, then adding the user
			if(nStatus == NERR_GroupNotFound)
			{
				EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Group %s does not exist, adding it"),currentGroup);
				localgroup_info.lgrpi1_name = currentGroup;
				localgroup_info.lgrpi1_comment = TEXT("pGina Created Group");

				nStatus = NetLocalGroupAdd(serverToUse,1,(LPBYTE)&localgroup_info,NULL); 

				if(nStatus == NERR_Success)
					nStatus = NetLocalGroupAddMembers(serverToUse,currentGroup,3,(LPBYTE)&memberInfo,1);
			}

			if(nStatus == NERR_Success)
				EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Added %s to group %s (global group)"),Username,currentGroup);
			else
				EventLog(true,EVENTLOG_ERROR_TYPE,TEXT("Adding %s to group %s failed (global group)"),Username,currentGroup);

			location = _tcstok(NULL,TEXT(";"));   
		}
	}

	if(fullList)
		free(fullList);

	if(settingsInfo.userGroups != NULL)
	{
		// First copy the whole list, so we dont screw with it, doing the plugin specified ones last
		fullList = _tcsdup(settingsInfo.userGroups);   

		// Pull of the first in the list   
		location = _tcstok(fullList,TEXT(";"));

		while(location != NULL)
		{
			currentGroup = _tcsdup(location);   

			nStatus = NetLocalGroupAddMembers(serverToUse,currentGroup,3,(LPBYTE) &memberInfo,1);

			// Try adding the group, then adding the user
			if(nStatus == NERR_GroupNotFound)
			{
				EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Group %s does not exist, adding it"),currentGroup);
				localgroup_info.lgrpi1_name = currentGroup;
				localgroup_info.lgrpi1_comment = TEXT("pGina Created Group");

				nStatus = NetLocalGroupAdd(serverToUse,1,(LPBYTE)&localgroup_info,NULL); 

				if(nStatus == NERR_Success)
					nStatus = NetLocalGroupAddMembers(serverToUse,currentGroup,3,(LPBYTE)&memberInfo,1);
			}

			if(nStatus == NERR_Success)
				EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Added %s to group %s"),Username,currentGroup);
			else
				EventLog(true,EVENTLOG_ERROR_TYPE,TEXT("Adding %s to group %s failed"),Username,currentGroup);

			location = _tcstok(NULL,TEXT(";"));   
		}   
	}

	if(fullList)
		free(fullList);

	if(Username)
		free(Username);

	if(manageDomain)
	{
		RevertToSelf();
		CloseHandle(hUser);
		if(serverToUse)
			free(serverToUse);
	}
} 

/**
 * Get the window title from the registry
 */
LPTSTR GetTitle(void)
{
    LPTSTR title = NULL;

    title = regReadString(_T("title"));

    if(windowTitle) {
        free(windowTitle);
    }

    if(title) {
        windowTitle = _tcsdup(title);
        free(title);
    } else {
        windowTitle = NULL;
    }

    return windowTitle;
}

/**
 * Get the motd from the registry
 */
LPTSTR GetMotd(void)
{
    LPTSTR motd = NULL;

    motd = regReadString(_T("motd"));

    if(motdText) {
        free(motdText);
    }

    if(motd) {
        motdText = _tcsdup(motd);
        free(motd);
    } else {
        motdText = NULL;
    }

    return motdText;
}

/**
 * Load the custom login logo and get a handle to it
 */
BOOL LoadLogo(void)
{
	if(customLogo != NULL)  {
		if(wcslen(customLogo) > 2) {
			if(logoBmp)
				DeleteObject(logoBmp);

			logoBmp = LoadImage((HINSTANCE)NULL,(LPCTSTR)customLogo,IMAGE_BITMAP,0,0,LR_LOADFROMFILE);

			return true;	
		} 
	} else {
		return false;
	}

	return false;
}

/**
 * Load the custom locked logo and get a handle to it
 */
BOOL LoadLockedLogo(void)
{
	if(customLocked != NULL)  {
		if(wcslen(customLocked) > 2) {
			if(locked_logoBmp)
				DeleteObject(locked_logoBmp);
	
			locked_logoBmp = LoadImage((HINSTANCE)NULL,(LPCTSTR)customLocked,IMAGE_BITMAP,0,0,LR_LOADFROMFILE);
	
			return true;	
		} 
	} else {
		return false;
	}

    return false;
}

/**
 * Determine whether we ware supposed to hide the plugin info
 */
BOOL HidePlugin(void)
{
    return regReadBool(_T("hidePlugin"));
}

/**
 * Determine whether we are supposed to hide the motd or not
 */
BOOL HideMotd(void)
{
    return regReadBool(_T("hideMotd"));
}

/**
 * Show a browse dialog for file selection
 */
LPTSTR browseForFile(int type)
{
	TCHAR szFile[30000] = { 0};
	TCHAR *dllFilter =TEXT("pGina Plugins (*.dll)\0") TEXT("*.dll\0") TEXT("All Files\0") TEXT("*.*\0");
	TCHAR *bmpFilter =TEXT("Bitmap Files (*.bmp)\0") TEXT("*.bmp\0") TEXT("All Files\0") TEXT("*.*\0");
	TCHAR *scriptFilter = TEXT("Executable Scripts (*.cmd)\0") TEXT ("*.cmd\0") 
						  TEXT("Batch Files (*.bat)\0") TEXT ("*.bat\0")
						  TEXT("Executable Programs (*.exe)\0") TEXT("*.exe\0")
						  TEXT("All Files\0") TEXT("*.*\0");
	TCHAR szCurDir[30000];
	OPENFILENAME ofn;

	GetCurrentDirectory(29999,szCurDir);
	ZeroMemory(&ofn,sizeof(OPENFILENAME));

	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hwndOwner = GetParent((HWND)pGinahInstance);

	if(type == PGINA_PLUGIN_DLL)
		ofn.lpstrFilter = dllFilter;
	else if(type == PGINA_LOGO_BMP)
		ofn.lpstrFilter = bmpFilter;
	else if(type == PGINA_SCRIPT_FILES)
		ofn.lpstrFilter = scriptFilter;

	ofn.nFilterIndex = 1;
	ofn.lpstrFile = szFile;
	ofn.nMaxFile = sizeof(szFile);
	ofn.lpstrTitle = _T("Open");
	ofn.nMaxFileTitle = sizeof (ofn.lpstrTitle);

	if(GetOpenFileName(&ofn))
		return _tcsdup(ofn.lpstrFile);
	else
		return NULL;
}

/**
 * Run the post login script
 */
void runPostLoginScript(PWSTR pszDesktopName)
{
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	WCHAR buffer[MAX_PATH];
	LPVOID pEnvBlock;

	if(UserLoginScript)
	{
		ZeroMemory( &si, sizeof(si) );
		ZeroMemory( &pi, sizeof(pi) );
		//ZeroMemory( &buffer,sizeof(buffer));
		si.cb = sizeof(si);
		si.wShowWindow = SW_HIDE;
		si.lpTitle = TEXT("Login Script");
		si.lpDesktop = pszDesktopName;

		if(!ImpersonateLoggedOnUser(hUser))
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Impersonation during mapping failed."));

		// expand %username% and %password% in the login script string
		// creating the environment block allows other vars to be expanded by windows
		ExpandEnvStrings(UserLoginScript, buffer, MAX_PATH);
		CreateEnvironmentBlock(&pEnvBlock, hUser, FALSE);
		if(CreateProcessAsUser(hUser,
						NULL, // Title
						buffer,
						(LPSECURITY_ATTRIBUTES)NULL,
						(LPSECURITY_ATTRIBUTES)NULL,
						FALSE,
						CREATE_UNICODE_ENVIRONMENT,//CREATE_NO_WINDOW,
						pEnvBlock, 
						NULL,
						&si,
						&pi))
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("CreateProcessAsUser returned true"));
		else
			EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("CreateProcessAsUser returned false, last error: %u"),GetLastError());

		DestroyEnvironmentBlock(pEnvBlock);
		RevertToSelf();
	}
}

/** 
 * Scramble the current users password into a 10 letter random password
 */
void scramblePassword(void)
{
	USER_INFO_1003 new_passwd;   
	NET_API_STATUS nStatus;
	LPTSTR tempPassword;
	LPTSTR domainToUse;
	LPTSTR serverToUse = NULL;
	HANDLE hUser;
	WCHAR newChar;
	int x = 0;

	// Determine what domain to do all this against
	if(!manageDomain)
		domainToUse = TEXT(".");	// Local machine
	else if(manageDomain)
	{
		if(settingsInfo.defaultDomain)
			domainToUse = settingsInfo.defaultDomain;
		else
			domainToUse	= defaultDomain;
	}

	// If we are to manage the domain, we must first login as a domain admin
	if(manageDomain)
	{
		// make sure we can login as the domain admin
		if(!LogonUser(adminUsername,domainToUse,adminPassword,LOGON32_LOGON_INTERACTIVE,LOGON32_PROVIDER_DEFAULT,&hUser))
		{
			// Login failed
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Logging in as an admin to domain %s with username %s failed."),domainToUse,adminUsername);
			return;
		}
		else
		{
			// Otherwise we are in, Impersonate this user...
			ImpersonateLoggedOnUser(hUser);
			serverToUse = getDNSfromDomain();
		}
	}

	tempPassword = _tcsdup(TEXT("0123456789"));

	for(;x < 10; x++)
	{
		newChar = (rand() % 1024);
		tempPassword[x] = newChar;
	}

	new_passwd.usri1003_password = tempPassword;

	nStatus = NetUserSetInfo(serverToUse,(LPCWSTR) settingsInfo.Username,1003,(LPBYTE) &new_passwd,NULL);

	if(nStatus != NERR_Success)
		EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Scrambling of password failed"));
	else
		EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Scrambled password"));

	if(manageDomain)
	{
		RevertToSelf();
		CloseHandle(hUser);
		if(serverToUse)
			free(serverToUse);
	}

}

/**
 * Get the DNS/Netbios name for the domain controller - Thanks to Ingo for the code! 
 */
LPTSTR getDNSfromDomain(void) 
{
	DWORD err;
	PDOMAIN_CONTROLLER_INFO DomainControllerInfo;
	LPTSTR domainToUse = NULL;

	// Determine the correct domain
	if(!manageDomain)
		return NULL;
	else
	{
		if(settingsInfo.defaultDomain)
			domainToUse = settingsInfo.defaultDomain;
		else
			domainToUse	= defaultDomain;
	}

	if((err = DsGetDcName(NULL, domainToUse, NULL, NULL, 0, &DomainControllerInfo)) == ERROR_SUCCESS)
	{
		EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("DsGetDcName yielded: %s"),DomainControllerInfo->DomainControllerName);
		return _tcsdup(DomainControllerInfo->DomainControllerName);
	}
	else
	{
		EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("DsGetDcName() failed: %d\n"), err);
		return NULL;
	}
}

/**
 * Get the version of the windows OS
 */
int getWindowsVersion(void)
{
	BOOL bOsVersionInfoEx;

	// Try calling GetVersionEx using the OSVERSIONINFOEX structure.
	// If that fails, try using the OSVERSIONINFO structure.

	ZeroMemory(&osvi, sizeof(OSVERSIONINFOEX));
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);

	if( !(bOsVersionInfoEx = GetVersionEx ((OSVERSIONINFO *) &osvi)) )
	{
		osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
		if(! GetVersionEx ( (OSVERSIONINFO *) &osvi) )
			return WINDOWS_NT_4; // Default to NT 4
	}

	switch(osvi.dwPlatformId)
	{
	// Test for the Windows NT product family.
	case VER_PLATFORM_WIN32_NT:

		// Test for the specific product family.
		if( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 2 )
			return WINDOWS_2003;

		if( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 1 )
			return WINDOWS_XP;

		if( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0 )
			return WINDOWS_2000;

		if( osvi.dwMajorVersion <= 4 )
			return WINDOWS_NT_4;

		break;

		// Test for the Windows 95 product family.
	case VER_PLATFORM_WIN32_WINDOWS:
		return WINDOWS_SUCKY;
		break;
	}
	memset(windowsBuild,'\0',2048);
	wsprintf(windowsBuild,_T("%d.%d %s (Build %d)"), osvi.dwMajorVersion, osvi.dwMinorVersion, osvi.szCSDVersion ,osvi.dwBuildNumber & 0xFFFF);
	// Default
	return WINDOWS_NT_4; // Default to NT 4
}

/**
 * Expand %% variables into a buffer
 */
void ExpandEnvStrings(TCHAR *original,TCHAR *dest,DWORD destsize)
{
	int length_source = 0,x=0;
	int dest_pos = 0;
	PTCHAR  pszSystemName;
	TCHAR *ip_address = NULL;
	TCHAR *mac_address = NULL;
	int start_macro = 0, end_macro = 0;

	pszSystemName = GlobalAlloc( GPTR, 2048 );
	memset(pszSystemName,'\0',2048);
	if( ! pszSystemName )
		return;

	memset(dest,'\0',destsize);
	GetEnvironmentVariable( _T("COMPUTERNAME"), pszSystemName, 2047 );
	ip_address = get_ip();
	mac_address = get_mac();

	length_source = wcslen(original);

	start_macro = 1;
	end_macro = 0;
	// Look at each char
	for(x=0;x<length_source;x++) {
		if(original[x] != L'%' && start_macro) {	// If we are NOT in a macro, just copy over
			dest[dest_pos++] = original[x];
		} else {
			// We are either in a macro, or need to start one
			if(original[x] == L'%' && start_macro) {
				start_macro = 0;
				end_macro = 1;

                // Figure out what macro this is, strcat the actual value onto dest
				if(wcsstr(original+x+1,L"user") == original+x+1) {
					if(settingsInfo.Username) {
						wcscat(dest,settingsInfo.Username);
						dest_pos += wcslen(settingsInfo.Username);
					}
				} else if(wcsstr(original+x+1,L"machine") == original+x+1) {
                    wcscat(dest,pszSystemName);
					dest_pos += wcslen(pszSystemName);
				} else if(wcsstr(original+x+1,L"ip") == original+x+1) {
					wcscat(dest,ip_address);
					dest_pos += wcslen(ip_address);
				} else if(wcsstr(original+x+1,L"mac") == original+x+1) {
					wcscat(dest,mac_address);
					dest_pos += wcslen(mac_address);
				} else if(wcsstr(original+x+1,L"password") == original+x+1) {
					if(settingsInfo.Password) {
						wcscat(dest,settingsInfo.Password);
						dest_pos += wcslen(settingsInfo.Password);
					}
				}
			} else if(original[x] == L'%' && end_macro){	// We hit hte end of a macro
				end_macro = 0;
				start_macro = 1;
			}
		}
	}


	if(ip_address) free(ip_address);
	if(mac_address) free(mac_address);
	GlobalFree(pszSystemName);
}

/**
	Locate the proper locale file to use, default to _en.txt if 
	one is not available for our current locale. Set locale_file
	to full path.
*/
void getLocaleFile(void) {
	DWORD err;
	HMODULE hmBinary;
	wchar_t lpLCData[10];
	wchar_t langShort[10];
	wchar_t psz[MAX_PATH], myPath[MAX_PATH], languagePath[MAX_PATH];
	wchar_t *pDest;
	int cchData = 10;

	memset(langShort,'\0',10);

	// determine default system locale and build path to resource DLL
	err = GetLocaleInfo(
					   LOCALE_SYSTEM_DEFAULT,	   // locale identifier
					   LOCALE_SABBREVLANGNAME,	   // information type
					   lpLCData,				   // information buffer
					   cchData					   // size of buffer
					   );

	if(err == 0)
	{
		err = GetLastError();
		EventLog(false,EVENTLOG_INFORMATION_TYPE,TEXT("Get Default system locale failed: %d"), err);  
	}
	else
	{
		EventLog(false,EVENTLOG_INFORMATION_TYPE,TEXT("Default system locale: %s"), lpLCData);
	}

	// We only want the first two characters from lpLCData, insert NULL one character before the end
	wcsncpy(langShort,lpLCData,2);

	// get path of this DLL
	hmBinary = GetModuleHandle(L"pgina.dll");
	GetModuleFileName(hmBinary, psz, MAX_PATH); 

	// build path to language DLL
	pDest = _tcsstr( _tcslwr(psz), L"pgina.dll");
	_tcsncpy(myPath, psz, pDest-psz);
	myPath[pDest-psz] = '\0';			// bad boy
	_stprintf(languagePath, TEXT("%slocale_%s.txt"), myPath, langShort);

	EventLog(false,EVENTLOG_INFORMATION_TYPE,TEXT("Locale Path: %s"), languagePath);

	// Verify the file exists
    if(GetFileAttributes(languagePath) == 0xffffffff)
	{
		// Default to english
		_stprintf(languagePath,TEXT("%slocale_en.txt"),myPath);
		EventLog(false,EVENTLOG_INFORMATION_TYPE,TEXT("Locale Path falling back on: %s"), languagePath);
	}

	locale_file = _tcsdup(languagePath);
}

/** Setup and send a troubleshooting report
 * 
 * @author Nate
 * @todo Show report questionaire and upload report
 */
void sendTroubleshootingReport(void)
{
	windowsVersion = getWindowsVersion();
	DialogBox(pGinahInstance,MAKEINTRESOURCE(IDD_TROUBLESHOOTING),NULL,(DLGPROC)TSBoxProc);
}

/** Actually submit (via libcurl) the troubleshooting report
 */
void SubmitReport(wchar_t *fullname, wchar_t *email, wchar_t *org, wchar_t *data, wchar_t *desc)
{
	CURL *curl;
	CURLcode res;

	struct HttpPost *formpost=NULL;
	struct HttpPost *lastptr=NULL;
	struct curl_slist *headerlist=NULL;
	char buf[] = "Expect:";
	char *c_fullname = convert_to_char(fullname);
	char *c_email = convert_to_char(email);
	char *c_org = convert_to_char(org);
	char *c_data = convert_to_char(data);
	char *c_desc = convert_to_char(desc);

	curl_global_init(CURL_GLOBAL_ALL);

	/* Fill in the file upload field */
	curl_formadd(&formpost,
				 &lastptr,
				 CURLFORM_COPYNAME, "report_text", // field name
				 CURLFORM_PTRCONTENTS, c_data, // set to
				 CURLFORM_END);

	curl_formadd(&formpost,
				 &lastptr,
				 CURLFORM_COPYNAME, "fullname",	// field name
				 CURLFORM_PTRCONTENTS, c_fullname, // set to
				 CURLFORM_END);

	curl_formadd(&formpost,
				 &lastptr,
				 CURLFORM_COPYNAME, "orgname", // field name
				 CURLFORM_PTRCONTENTS, c_org, // set to
				 CURLFORM_END);

	curl_formadd(&formpost,
				 &lastptr,
				 CURLFORM_COPYNAME, "email", // field name
				 CURLFORM_PTRCONTENTS, c_email,	// set to
				 CURLFORM_END);

	curl_formadd(&formpost,
				 &lastptr,
				 CURLFORM_COPYNAME, "desc",	// field name
				 CURLFORM_PTRCONTENTS, c_desc, // set to
				 CURLFORM_END);

	curl = curl_easy_init();
	/* initalize custom header list (stating that Expect: 100-continue is not
	   wanted */
	headerlist = curl_slist_append(headerlist, buf);
	if(curl)
	{
		/* what URL that receives this POST */
		curl_easy_setopt(curl, CURLOPT_URL, "http://www.xpasystems.com/bugreport/report.php");
		curl_easy_setopt(curl, CURLOPT_HTTPHEADER, headerlist);
		curl_easy_setopt(curl, CURLOPT_HTTPPOST, formpost);
		res = curl_easy_perform(curl);

		/* always cleanup */
		curl_easy_cleanup(curl);

		/* then cleanup the formpost chain */
		curl_formfree(formpost);
		/* free slist */
		curl_slist_free_all (headerlist);

		if(res)
		{
			//pGinaMessageBox(IDS_CURLERROR,0,MB_ICONERROR | MB_OK);
			MessageBox(GetForegroundWindow(),TranslatedString(GSECTION,TEXT("IDS_CURLERROR")),NULL,MB_ICONERROR | MB_OK);
		}
	}

	GlobalFree(c_email);
	GlobalFree(c_fullname);
	GlobalFree(c_org);
	GlobalFree(c_data);
	GlobalFree(c_desc);
}

/**
 * Convert a wide char to a single byte char array
 */
char * convert_to_char(const TCHAR *from)
{
	char *to = (char *) malloc(wcslen(from)*sizeof(TCHAR)+1);

	WideCharToMultiByte(CP_ACP,
						0,
						(LPCWSTR)from,
						-1,
						to,
						wcslen(from) *sizeof(TCHAR)+1,
						NULL,
						NULL);

	return to;
}

/**
 * Attempt to auth the user in global pGinaInfo *settingsInfo
 */
BOOL CheckAdminUnlock(LPTSTR Username, LPTSTR Password)
{
	LPUSER_INFO_2 userInfo = NULL;
    DWORD level = 2;
	LPTSTR domainToUse = NULL;
	LPTSTR serverToUse = NULL;
	BOOL userIsSame = FALSE;
	pGinaInfo mySettingsInfo;


	if(wcsncmp(Username,settingsInfo.Username,wcslen(Username)) == 0 && wcslen(Username) == wcslen(settingsInfo.Username)) {
		userIsSame = TRUE;
	}

	// First try authenticating via plugin
	mySettingsInfo.userGroups = NULL;
	mySettingsInfo.userFullName = NULL;
	mySettingsInfo.userDescription = NULL;
	mySettingsInfo.isAdmin = false;
	mySettingsInfo.allowPassChange = true;
	mySettingsInfo.errorString = NULL;
	mySettingsInfo.defaultDomain = NULL;
	mySettingsInfo.Username = _tcsdup(Username);
	mySettingsInfo.Password = _tcsdup(Password);

	if(!allowSpaces)
	{
		// Check for spaces in the username
		if(wcsstr(mySettingsInfo.Username,TEXT(" ")))
		{
			// There are spaces, sorry - none of that allowed!
			return false;
		}
	}

	// Try login via Plugin
	if(pluginUserLogin(mySettingsInfo.Username,mySettingsInfo.Password,&mySettingsInfo))
	{
		// User is authenticated
		mySettingsInfo.authType = AUTH_PLUGIN;

		// Make sure they are an admin
		if(mySettingsInfo.isAdmin || userIsSame) {
			return true;
		}
	}

	// If the plugin is required double check to ensure they arent an admin, and die
	if(isRequired)
	{
		if(mySettingsInfo.errorString != NULL)
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Required Plugin Error is %s"),mySettingsInfo.errorString);
		else
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Required Plugin Failed - no error provided."));

		if(manageDomain)
			serverToUse = getDNSfromDomain();
		else
			serverToUse	= NULL;

		// Make sure they arent trying to use a local or domain admin account, for which we must continue
		if(NetUserGetInfo(serverToUse,mySettingsInfo.Username,3,(LPBYTE *)&userInfo) == NERR_Success)
		{
			if(userInfo->usri2_priv != USER_PRIV_ADMIN )
			{
				EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("User is not an administrator, so failing."));
				return false;
			}
			else
				EventLog(false,EVENTLOG_INFORMATION_TYPE,TEXT("Administrator access allowed though required plugin failed."));
		}
		else
		{
			return false;
		}
	}

	EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Plugin authentication failed, moving on to Domain."));

	// See if we should try via a domain
	if(includeDomain)
	{
		// mySettingsInfo.defaultDomain takes precedence if set
		if(mySettingsInfo.defaultDomain)
			domainToUse = mySettingsInfo.defaultDomain;
		else
			domainToUse	= defaultDomain;

		EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Trying with domain: %s"),domainToUse);

		// Try the login via the domain
		if(LogonUser(mySettingsInfo.Username,domainToUse,mySettingsInfo.Password,LOGON32_LOGON_INTERACTIVE,LOGON32_PROVIDER_DEFAULT, &hUser))
		{
			CloseHandle(hUser);
			mySettingsInfo.authType = AUTH_DOMAIN;
			EventLog(false,EVENTLOG_AUDIT_SUCCESS,TEXT("User %s logged in via Domain (%s) successfully."),mySettingsInfo.Username,domainToUse);

			if(mySettingsInfo.errorString != NULL && displayErrors)
			{
				MessageBox(NULL,mySettingsInfo.errorString,TEXT("pGina Plugin Message"),MB_ICONERROR | MB_OK | MB_SYSTEMMODAL);
				EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Error returned by plugin during authentication attempt: %s"),mySettingsInfo.errorString);
			}

			// If they are an admin, return true
			if(NetUserGetInfo(domainToUse,mySettingsInfo.Username,3,(LPBYTE *)&userInfo) == NERR_Success)
			{
				if(userInfo->usri2_priv == USER_PRIV_ADMIN  || userIsSame)
				{
					return true;
				}
			}
		}
	}

	// Then try local SAM
	EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Domain authentication failed, moving on to SAM."));

	if(LogonUser(mySettingsInfo.Username,TEXT("."),mySettingsInfo.Password,LOGON32_LOGON_INTERACTIVE,LOGON32_PROVIDER_DEFAULT, &hUser))
	{
		CloseHandle(hUser);
		mySettingsInfo.authType = AUTH_LOCAL;
		EventLog(false,EVENTLOG_AUDIT_SUCCESS,TEXT("User %s logged in via local SAM successfully."),mySettingsInfo.Username);

		if(mySettingsInfo.errorString != NULL && displayErrors)
		{
			MessageBox(NULL,mySettingsInfo.errorString,TEXT("pGina Plugin Message"),MB_ICONERROR | MB_OK | MB_SYSTEMMODAL);
			EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Error returned by plugin during authentication attempt: %s"),mySettingsInfo.errorString);
		}

		// If they are an admin, return true
		if(NetUserGetInfo(NULL,mySettingsInfo.Username,3,(LPBYTE *)&userInfo) == NERR_Success)
		{
			if(userInfo->usri2_priv == USER_PRIV_ADMIN  || userIsSame)
			{
				return true;
			}
		}
	}

	// We can display an error here... or after the SAM check
	if(mySettingsInfo.errorString != NULL)
	{
		MessageBox(NULL,mySettingsInfo.errorString,TEXT("pGina Login Failure"),MB_ICONERROR | MB_OK | MB_SYSTEMMODAL);
		EventLog(false,EVENTLOG_ERROR_TYPE,TEXT("Error returned by plugin during authentication attempt: %s"),mySettingsInfo.errorString);
	}
	else {
		MessageBox(GetForegroundWindow(),TranslatedString(GSECTION,TEXT("IDS_ACCESSDENIED")),TranslatedString(GSECTION,TEXT("IDS_T_ACCESSDENIED")),MB_ICONERROR | MB_OK | MB_SYSTEMMODAL);
	}
		// "Access Denied.  Please check your username and password.  If the problem persists - contact the System Administrator"
		//pGinaMessageBox(IDS_ACCESSDENIED,IDS_T_ACCESSDENIED,MB_ICONERROR | MB_OK | MB_SYSTEMMODAL);

	EventLog(false,EVENTLOG_AUDIT_FAILURE,TEXT("Authentication of %s failed."),mySettingsInfo.Username);

	return false;
}


/**
 * Dup a string with LocalAlloc (for Wlx usage)
 */
PWSTR DupString(PWSTR pszString)
{
  DWORD cbString;
  PWSTR pszNewString;

  cbString = (wcslen(pszString) + 1) * sizeof(WCHAR);
  pszNewString = LocalAlloc(LMEM_FIXED, cbString);
  if (pszNewString)
  {
    CopyMemory(pszNewString, pszString, cbString);
  }
  return(pszNewString);
}

/**
 *	@brief Creates a duplicate copy of a string. The returned
 *		string is always in UNICODE format.
 *	@author Eric Parsons
 *
 *	Uses LocalAlloc to allocate memory for the new string. 
 * 	The out parameters for the Wlx* functions must be allocated
 *	this way and NOT using malloc or _tcsdup (which uses malloc).
 *	If the resulting string isn't passed back to WinLogon, then
 *	it must be freed with LocalFree().
 */
LPWSTR wlxsdup(LPTSTR str) {
	LPWSTR dup;
	int size = lstrlen(str) + 1;

	if (NULL == str)
		return NULL;

	dup = LocalAlloc(LMEM_FIXED, size * sizeof(WCHAR));
	if (dup) {
#ifdef UNICODE
		lstrcpy(dup, str);
#else
		MultiByteToWideChar(CP_ASC, MB_ERR_INVALID_CHARS, str, -1, dup, size);
#endif
	}
	return dup;
}

/**
	Get the translated string.  Checks the given section, as well
	as the Global section (in that order) and returns NULL if not found
*/
LPTSTR TranslatedString(LPTSTR Section, LPTSTR Key) {
	static LPTSTR retVal = NULL;
	DWORD GPPS = 0;
	DWORD stringMax = 0;

	if(retVal != NULL) {	// If we've been called before, free up memory
		free(retVal);
	}
	
	stringMax = 10000 * sizeof(TCHAR);
	retVal = (LPTSTR) malloc(stringMax);

	GPPS = GetPrivateProfileString(Section,Key,TEXT("NOTFOUND"),retVal,stringMax,locale_file);

	if(GPPS) {
		if(wcscmp(retVal,TEXT("NOTFOUND")) == 0) {
			// String wasnt found in the given spot, see if its in the Global section
			GPPS = GetPrivateProfileString(TEXT("Global"),Key,TEXT("NOTFOUND"),retVal,stringMax,locale_file);
			if(GPPS) {
				if(wcscmp(retVal,TEXT("NOTFOUND")) == 0) {
					free(retVal); // Nothin found, dont waste mem
					retVal = NULL;
				}
			}
		}
	}

	// At this point, retVal has our value of choice..
	return retVal;
}

/**
 * Check for the auto-login registry items, if they exist, and LogonUser verifies
 *   the credentials, then return true.  Otherwise, return false.
 */
BOOL isAutoAdminValid(void) {
	HKEY    hKey;
	LONG    lResult;
	TCHAR 	result[2048];
	DWORD 	dwBytesRead = 2048;
	DWORD	size = sizeof(DWORD);
	DWORD	AutoAdminLogon = 0, ForceAutoLogon = 0;
	TCHAR  *DefaultUsername = NULL, *DefaultPassword = NULL, *DefaultDomainName = NULL;
	BOOL retval = false;
	TCHAR  pszSystemName[2048];
	HANDLE hUser = NULL;

	GetEnvironmentVariable( _T("COMPUTERNAME"), pszSystemName, 2048 );

	lResult = RegOpenKeyEx(HKEY_LOCAL_MACHINE,_T("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon"),0, KEY_QUERY_VALUE, &hKey);

	if(lResult == ERROR_SUCCESS)
	{
		// AutoAdminLogon or ForceAutoLogon must be 1
		lResult = RegQueryValueEx(hKey,_T("AutoAdminLogon"),NULL,NULL,(LPBYTE)&AutoAdminLogon,&size);
		size = sizeof(DWORD);
		lResult = RegQueryValueEx(hKey,_T("ForceAutoLogon"),NULL,NULL,(LPBYTE)&ForceAutoLogon,&size);

		// DefaultUsername
        lResult = RegQueryValueEx(hKey,_T("DefaultUserName"),NULL,NULL,(LPBYTE)&result,&dwBytesRead);
		if(dwBytesRead > 0 && lResult == ERROR_SUCCESS) {
			DefaultUsername = _tcsdup(result);
		}
		dwBytesRead = 2048;
		// DefaultPassword
		lResult = RegQueryValueEx(hKey,_T("DefaultPassword"),NULL,NULL,(LPBYTE)&result,&dwBytesRead);
		if(dwBytesRead > 0 && lResult == ERROR_SUCCESS) {
			DefaultPassword = _tcsdup(result);
		}
		dwBytesRead = 2048;
		// DefaultDomainName
		lResult = RegQueryValueEx(hKey,_T("DefaultDomainName"),NULL,NULL,(LPBYTE)&result,&dwBytesRead);
		if(dwBytesRead > 0 && lResult == ERROR_SUCCESS) {
			DefaultDomainName = _tcsdup(result);
		}

		lResult= RegFlushKey(hKey);
        RegCloseKey(hKey);

		// Check and see if all is kosher
		if(AutoAdminLogon != 1 && ForceAutoLogon != 1) goto Cleanup; // One of these must be set
		if(DefaultUsername == NULL || DefaultPassword == NULL || DefaultDomainName == NULL) goto Cleanup;
		if(wcscmp(DefaultDomainName,pszSystemName) != 0) goto Cleanup; // Must be for the local machine

        // We have values for everything, now make sure we can login
        if(LogonUser(DefaultUsername,DefaultDomainName,DefaultPassword,LOGON32_LOGON_INTERACTIVE,LOGON32_PROVIDER_DEFAULT, &hUser)) {
			CloseHandle(hUser);
			// Worked, so set our val as true
			retval = true;
		} 

		Cleanup:
			if(DefaultUsername) free(DefaultUsername);
			if(DefaultPassword) free(DefaultPassword);
			if(DefaultDomainName) free(DefaultDomainName);
			return retval;
		
	} else {
		return false;
	}
}

/**
 * Get the machines ip address
 */
TCHAR *get_ip(void) {
	PIP_ADAPTER_INFO pAdapterInfo;
	PIP_ADAPTER_INFO pAdapter = NULL;
	DWORD dwRetVal = 0;
	ULONG ulOutBufLen = 0;
	TCHAR *result = NULL;

	pAdapterInfo = (IP_ADAPTER_INFO *) malloc( sizeof(IP_ADAPTER_INFO) );
	ulOutBufLen = sizeof(IP_ADAPTER_INFO);

	// Make an initial call to GetAdaptersInfo to get
	// the necessary size into the ulOutBufLen variable
	if (GetAdaptersInfo( pAdapterInfo, &ulOutBufLen) == ERROR_BUFFER_OVERFLOW) {
		free(pAdapterInfo);
		pAdapterInfo = (IP_ADAPTER_INFO *) malloc (ulOutBufLen); 
	}

	if ((dwRetVal = GetAdaptersInfo( pAdapterInfo, &ulOutBufLen)) == NO_ERROR) {
		pAdapter = pAdapterInfo;
		if(pAdapter) {
			result = convert_to_wchar(pAdapter->IpAddressList.IpAddress.String);
			free(pAdapterInfo);
			return result;
		}
	} else {
		return _tcsdup(_T("[Unknown IP Address]"));
	}
	return _tcsdup(_T("[Unknown IP Address]"));
}

/**
 * Get the machines mac address
 */
TCHAR *get_mac(void) {
	PIP_ADAPTER_INFO pAdapterInfo;
	PIP_ADAPTER_INFO pAdapter = NULL;
	DWORD dwRetVal = 0;
	ULONG ulOutBufLen = 0;
	TCHAR *result = NULL;
	TCHAR buff[100];
	int x = 0;

	pAdapterInfo = (IP_ADAPTER_INFO *) malloc( sizeof(IP_ADAPTER_INFO) );
	ulOutBufLen = sizeof(IP_ADAPTER_INFO);

	// Make an initial call to GetAdaptersInfo to get
	// the necessary size into the ulOutBufLen variable
	if (GetAdaptersInfo( pAdapterInfo, &ulOutBufLen) == ERROR_BUFFER_OVERFLOW) {
		free(pAdapterInfo);
		pAdapterInfo = (IP_ADAPTER_INFO *) malloc (ulOutBufLen); 
	}

	if ((dwRetVal = GetAdaptersInfo( pAdapterInfo, &ulOutBufLen)) == NO_ERROR) {
		pAdapter = pAdapterInfo;
		if(pAdapter) {
			// MAC address is no in pAdapter->Address, its BYTE array of pAdapter->AddressLength bytes
			if(pAdapter->Address && pAdapter->AddressLength) {
				result = (TCHAR *)malloc(sizeof(TCHAR) * pAdapter->AddressLength * 10); // make sure we have plenty of room
				memset(result,'\0',(sizeof(TCHAR) * pAdapter->AddressLength * 10));
				
				for(x=0;x<pAdapter->AddressLength;x++) {
					if( x==0) {
						wsprintf(result,L"%02X:",pAdapter->Address[x]);
					} else if(x<pAdapter->AddressLength-1 && x > 0) {
						wsprintf(buff,L"%02X:",pAdapter->Address[x]);
						wcscat(result,buff);
					} else {
						wsprintf(buff,L"%02X",pAdapter->Address[x]);
						wcscat(result,buff);
					}
				}
			}

            free(pAdapterInfo);
			return result;
		}
	} else {
		return _tcsdup(_T("[Unknown IP Address]"));
	}
	return _tcsdup(_T("[Unknown IP Address]"));
}



/**
 * Convert a single byte char to wide char array
 */
TCHAR * convert_to_wchar(const char *from)
{
	TCHAR *to;
	if (from == NULL)
		return NULL;

	to = (TCHAR *) malloc((strlen(from)+1)*sizeof(TCHAR));
	MultiByteToWideChar(CP_ACP,
						0,
						(LPCSTR)from,
						strlen(from)+1,
						(LPWSTR)to,
						(strlen(from)+1)*sizeof(TCHAR));                    

	return to;
}

/*
    An encryption key is loaded from a file called pGinaKey.dll that must 
    reside in %SYSTEM32%.  If it can't be loaded, or a NULL key is returned,
	encryption will still happen, but anyone on the machine can decrypt the values
	with a NULL key.
*/
void GetEncryptionKey(void) {
	HANDLE encryptionDLL = NULL;
    DATA_BLOB *keyFromDLL = NULL;
	encryptionKey.pbData = NULL;
	encryptionKey.cbData = 0;

	// Make sure its a DLL and set the handle
	if(!(encryptionDLL = LoadLibrary(_T("pGinaKey.dll"))))
	{
   		EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Loaded encryption key dll failed, defaulting to NULL key"));
		return;
	}

	// Get a handle to the variable we want
	dllEncryptionKey = (DLLGETENCRYPTIONKEY) GetProcAddress(encryptionDLL,"EncryptionKey");

	// If unable to load
	if(dllEncryptionKey == NULL) {
    	EventLog(true,EVENTLOG_INFORMATION_TYPE,TEXT("Loaded encryption key dll succeeded, but loading key failed"));
		return;
	}

    keyFromDLL = dllEncryptionKey();

    if(keyFromDLL) {
        // Get the size of the key
        encryptionKey.cbData = keyFromDLL->cbData;
        encryptionKey.pbData = (BYTE *)malloc(keyFromDLL->cbData); 

        // Copy the key in
        memcpy(encryptionKey.pbData,keyFromDLL->pbData,keyFromDLL->cbData);
    }

	FreeLibrary(encryptionDLL);              
}

/**
 * Encrypt a string using CryptProtectData()
 */
LPTSTR encryptString(LPTSTR origString) {
	DATA_BLOB original;
	DATA_BLOB encrypted;
	DATA_BLOB encrypted_encoded;
	char *as_char = NULL;

	if(origString == NULL || wcslen(origString) < 2) {
		return origString;
	}

	original.cbData = (DWORD)wcslen(origString) * sizeof(TCHAR) + sizeof(TCHAR);
	original.pbData = (BYTE *)origString;


	if(CryptProtectData(&original,_T("pGina Setting String"),&encryptionKey,NULL,NULL, CRYPTPROTECT_LOCAL_MACHINE | CRYPTPROTECT_UI_FORBIDDEN, &encrypted)) {
		encode(&encrypted,&encrypted_encoded); // Encode the binary value as a string
		return (LPTSTR)encrypted_encoded.pbData;
	} else {
		EventLog(false,EVENTLOG_ERROR_TYPE,_T("Error encrypting data\n"));
		return origString;	// We return the original string
	}
}

/**
 * Decrypt a string using CryptUnprotectData()
 */
LPTSTR decryptString(LPTSTR origString) {
	DATA_BLOB original;
	DATA_BLOB original_decoded;
	DATA_BLOB decrypted;
    LPTSTR description = _T("pGina Setting String");

	if(origString == NULL || wcslen(origString) < 2) {
		return origString;
	}

	original.cbData = (DWORD)wcslen(origString) * sizeof(TCHAR) + sizeof(TCHAR);
	original.pbData = (BYTE *)origString;

	decode(&original,&original_decoded);

	if(CryptUnprotectData(&original_decoded,&description,&encryptionKey,NULL,NULL, CRYPTPROTECT_LOCAL_MACHINE | CRYPTPROTECT_UI_FORBIDDEN, &decrypted)) {
		return (LPTSTR)_tcsdup(decrypted.pbData);
	} else {
		EventLog(false,EVENTLOG_ERROR_TYPE,_T("Error decrypting data\n"));
		return NULL;	// We return NULL
	}
}

/**
 * Encode a binary data blob as text (hex)
 */
void encode(DATA_BLOB *data,DATA_BLOB *result) {
	char temp[3];
	char *buffer;
	int x = 0;

	buffer = (char *)malloc((data->cbData * 2) + 1);
	memset(buffer,0,((data->cbData * 2) + 1));

	for(x=0;x<(int)data->cbData;x++) {
		sprintf(temp,"%02x",data->pbData[x]);
		strcat(buffer,temp);
	}

	result->cbData = (DWORD)strlen(buffer)*sizeof(TCHAR) + sizeof(TCHAR);
	result->pbData = (BYTE *)convert_to_wchar(buffer);
	free(buffer);
}

/**
 * The reverse of encode() - decodes a string of hex values
 *   into a binary data blob.
 */
void decode(DATA_BLOB *data,DATA_BLOB *result) {
	long temp = 0;
	char *as_char = NULL;
	char *buffer = NULL;
	int x = 0,y =0;
	char tempc[3];
	tempc[2] = '\0';


	as_char = convert_to_char((TCHAR *)data->pbData);
	buffer = (char *)malloc(strlen(as_char)+1);
	memset(buffer,0,(strlen(as_char)+1));

	for(x=0;x<(int)strlen(as_char);x+=2) {
		tempc[0] = as_char[x];
		tempc[1] = as_char[x+1];
		buffer[y++] = (char)strtol(tempc,'\0',16);	
	}

	result->cbData = (DWORD)strlen(as_char)+1; 
	result->pbData = (BYTE *)buffer;
	free(as_char);
}

/**
 * Extracts a user's SID from a security token. 
 * The returned SID structure must be freed with LocalFree.
 */
PSID GetUserSid(HANDLE hUser) {
	PTOKEN_USER ptu;
	DWORD size;
	PSID sid = NULL;
	
	// This first call gets the size needed for the SID
	GetTokenInformation(hUser, TokenUser, NULL, 0, &size);
	ptu = (PTOKEN_USER)LocalAlloc(LMEM_FIXED, size);

	if (GetTokenInformation(settingsInfo.hUser, TokenUser, ptu, size, &size)) {
		size = GetLengthSid(ptu->User.Sid);
		sid = (PSID)LocalAlloc(LMEM_FIXED, size);
		if (!CopySid(size, sid, ptu->User.Sid)) {
			LocalFree(sid);
			sid = NULL;
		}
	}
	
	LocalFree(ptu);
	return sid;
}

/**
 * Opens a user's registry key. This function should be used instead of
 * HKEY_CURRENT_USER since the latter will usually refer to the registry key
 * for the system account in most of the GINA functions.
 * This function will fail if the profile for the user hasn't been loaded.
 * The returned key MUST be closed with RegCloseKey, otherwise the profile will
 * be locked open when the user logs out and many strange things will happen...
 */
HKEY RegOpenUser(HANDLE hUser, REGSAM desiredAccess) {
	PSID sid;
	LPTSTR strSid;
	HKEY key = NULL;

	sid = GetUserSid(hUser);
	if (ConvertSidToStringSid(sid, &strSid)) {
		RegOpenKeyEx(HKEY_USERS, strSid, 0, desiredAccess, &key);
		LocalFree(strSid);
	}

	LocalFree(sid);
	return key;
}

BOOL RegAddKeyPermission(HKEY key, PSID user, DWORD permissions) {
	EXPLICIT_ACCESS ace;
	PACL oldACL, newACL;
	BOOL ret = FALSE;

	if (ERROR_SUCCESS == GetSecurityInfo(key, SE_REGISTRY_KEY, DACL_SECURITY_INFORMATION, 
		NULL, NULL, &oldACL, NULL, NULL))
	{
		// Create and add an ACE to the user's registry with required permissions for the user 
		BuildExplicitAccessWithName(&ace, settingsInfo.Username, permissions, SET_ACCESS,
			SUB_CONTAINERS_AND_OBJECTS_INHERIT);
		if (ERROR_SUCCESS == SetEntriesInAcl(1, &ace, oldACL, &newACL)) {
			ret = (ERROR_SUCCESS == SetSecurityInfo(key, SE_REGISTRY_KEY, 
				DACL_SECURITY_INFORMATION, NULL, NULL, newACL, NULL));
			LocalFree(newACL);
		}
		LocalFree(oldACL);
	}

	return ret;
}

BOOL RegRemoveKeyPermission(HKEY key, PSID user) {
	DWORD i;
	PACL acl;
	BOOL ret = FALSE;
	PSID sid;
	PACE_HEADER ace = NULL;
	
	if (ERROR_SUCCESS == GetSecurityInfo(key, SE_REGISTRY_KEY, DACL_SECURITY_INFORMATION, 
			NULL, NULL, &acl, NULL, NULL))
	{
		// Delete any ACE that grants access to the user
		for (i = 0; i < acl->AceCount; i++) {
			if (GetAce(acl, i, &ace) && ace->AceType == ACCESS_ALLOWED_ACE_TYPE)
			{
				sid = (PSID) &(((PACCESS_ALLOWED_ACE)ace)->SidStart);
				if (EqualSid(sid, user)) {
					DeleteAce(acl, i);
				}
			}
		}
		ret = (ERROR_SUCCESS == SetSecurityInfo(key, SE_REGISTRY_KEY, 
			DACL_SECURITY_INFORMATION, NULL, NULL, acl, NULL));
		LocalFree(acl);
	}

	return ret;
}
