/**
	@file 		pGina.h 
	@brief 		Core header file for the project
	@author 	Nathan Yocom
*/
/*
pGina - A modular GINA Interface for Windows
Copyright (C) 2002 Nathan Yocom, Michael Wright
			  UNICODE Conversions by Andrew Lansford

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email: nate.yocom@xpasystems.com
Web: http://pgina.xpasystems.com
Snail Mail:
  Nathan Yocom 
  9 Evergreen Farms Rd.
  Scarborough, ME 04074
  Phone: 207-450-4948
*/

#ifndef __PGINA_H__
#define __PGINA_H__
  
// All the common includes
#include <stdio.h>
#include <lm.h>
#include <wingdi.h>
#include <time.h>
#include <windows.h>
#include <winwlx.h>
#include <tchar.h>
#include <wchar.h>
#include <lmaccess.h>
#include <shellapi.h>
#include <Iphlpapi.h>
#include <Wincrypt.h>
#include <aclapi.h>

/** pGina version information */    
#define VERSION TEXT("pGina Version 1.8.4 - Nathan Yocom, Michael Wright - http://www.xpasystems.com")
#define VERSION_SHORT TEXT("1.8.4")

/** Define to export any given function */  
#define EXTERNAL_API __declspec(dllexport)  

/** Define to import any given function */
#define PLUGIN_API __declspec(dllimport)

/** Set true and false up so we can use them */
#define false 0
#define true 1

/** Flag to indicate authentication took place via a pGina plugin */
#define AUTH_PLUGIN 100

/** Flag to indicate authentication took place via the local SAM */
#define AUTH_LOCAL 101

/** Flag to indicate authenticate took place via the defaultDomain or settingsInfo.defaultDomain */                       
#define AUTH_DOMAIN 102

/** Flag indicates the user hit the cancel button on the login form */
#define LOGIN_CANCELED 10102

/** Flag indicates the user hit the Shutdown/Reboot button on the login form */
#define LOGIN_SHUTDOWNREBOOT 10103

/** Flag indicates the login attempt failed */
#define LOGIN_FAILED 10104

#define IDLE_LOGOFF 1000
#define IDLE_SCREENSAVER 1001
#define IDLE_CANCEL 1002
#define IDLE_LOCK 1003
#define IDLE_UNLOCK 1004
#define IDLE_MESSAGE_TIMEOUT WM_USER+10

#define GSECTION TEXT("Global")

/**
 *  @struct pGinaInfo pgina.h 
 *	@brief 	Structure used mostly for data passing between pGina and its plugins
 * 
 *  @param  pathMSGina      Full path to the Microsoft (c) GINA library DLL
 *  @param  pathPlugin      Full path to the pGina plugin that pGina will load
 *  @param  pathProfile     If non-NULL, Windows will try to load the user's profile from
 *                              this path. Note that this path CANNOT make reference to mapped drives
 *                              the path must be a local path or a UNC path. Can contain variables
 *                              like %username%
 * 
 *  @param  mapPaths        Semicolon seperated list of drives/paths to be mapped in the form:
 *                              DRIVE:\\PATH, such as: X:\\somehost\%username%;Z:\\anotherhost\path
 *  @param  Username        The username provided by the user to authenticate
 *  @param  Password        The password the user entered
 *  @param  Reserved        Currently Not Implemented - Reserved
 *  @param  isAdmin         Boolean value allowing the plugin to specify whether a user
 *                              should be an admin or not
 *  @param  disable         Boolean value that indicates pGina should act in passthrough mode
 *  @param  authType        Flag used to identify how the user was authenticated, can be AUTH_PLUGIN, AUTH_LOCAL 
 *  @param  hUser           Handle to the users security context in-case the plugin needs it
 *  @param  userGroups      Local groups to add the user to
 *  @param  userDescription Text to populate user's description with
 *  @param  userFullName    Text to populate user's Name field with
 *  @param  allowPassChange Should the Change Password button be allowd for this user
 *  @param  errorString     An error string returned by a plugin as a result of the UserLogin call
 *  @param  defaultDomain   A plugin specified Domain name for domain interaction (superceeds default Domain)
 *  @param  pathProfileDefault  If no profile is found in profilePath, Winlogon will copy one from this path 
 *                                  (otherwise, the local "Default User" profile will be copied). Format is the same as
 *                                  pathProfile (i.e. no mapped drive referrences)
 *  @param  sessionTimeout  Number of seconds a user is allowed to be logged in before they are locked out again, 0 == forever
 * 
*/
typedef struct pGinaInfo {
        LPTSTR pathMSGina;      
        LPTSTR pathPlugin;      
        LPTSTR pathProfile;
		LPTSTR mapPaths;
		LPTSTR Username;
		LPTSTR Password;
		LPTSTR Reserved;
		BOOL isAdmin;
		BOOL disabled;
		int authType;
		HANDLE hUser;
		LPTSTR userGroups;
		LPTSTR userDescription;
		LPTSTR userFullName;
		BOOL allowPassChange;
		LPTSTR errorString;
		LPTSTR defaultDomain;
		LPTSTR pathProfileDefault; 
        DWORD sessionTimeout; 
} pGinaInfo ;



// EXPORTED FUNCTIONS
EXTERNAL_API BOOL pGinaConfigure(void);


// INTERNAL UTILITY FUNCTIONS
BOOL LoadPlugin(void);                              // Load plugin of choice
BOOL MapDrives(void);                               // Map network drives
void SaveSettings(void);                            // Save all our settings to the registry
void LoadSettings(void);                            // Load all our settings from the registry
void EventLog(BOOL,WORD,TCHAR *,...);               // Log a string to the event log
BOOL CreateLocalAccount(void);
void DeleteLocalAccount(BOOL);
LPTSTR GetTitle(void);
LPTSTR GetMotd(void);
BOOL LoadLogo(void);
BOOL HidePlugin(void);
BOOL HideMotd(void);
void ExpandEnvStrings(TCHAR *,TCHAR *,DWORD);
BOOL CheckAdminUnlock(LPTSTR,LPTSTR);
BOOL AuthUser(LPTSTR, LPTSTR);
BOOL LoadLockedLogo(void);

BOOL isAutoAdminValid(void);
void SubmitReport(wchar_t *fullname, wchar_t *email, wchar_t *org, wchar_t *data,wchar_t *desc);
void DeleteFolder(LPCTSTR);
void dumpState(wchar_t *,unsigned int);
void AddUserToGroups(void);  

void runPostLoginScript(PWSTR);
void scramblePassword(void);
LPTSTR getDNSfromDomain(void);
int getWindowsVersion(void);
void getLocaleFile(void);
void sendTroubleshootingReport(void);
TCHAR * get_ip(void);
TCHAR *get_mac(void);
PSID GetUserSid(HANDLE hUser);
void DeMapDrives(void);


// CALLBACKS AND DIALOGS
BOOL CALLBACK PasswordBoxProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK LockedBoxProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK LockedLoginBoxProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam); 
BOOL CALLBACK TSBoxProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK LoginBoxProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);                  
BOOL CALLBACK ConfigGeneralDialogProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK ConfigLoginDialogProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK ConfigLockedDialogProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK ConfigAccountsDialogProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK ConfigDomainDialogProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK ConfigAdvancedDialogProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK ConfigAboutDialogProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK ConfigProfileDialogProc(HWND, UINT, WPARAM, LPARAM);
BOOL CALLBACK IdleBoxProc(HWND, UINT, WPARAM, LPARAM);
VOID CenterWindow(HWND hwnd);
LPTSTR browseForFile(int);
HRESULT MakeBold(HWND hWnd, BOOL fSize, LONG lfWeight);
HRESULT ShowConfigDialog(void);

#define PGINA_PLUGIN_DLL 1
#define PGINA_LOGO_BMP 2
#define PGINA_SCRIPT_FILES 3


// REGISTRY HANDLING
TCHAR * regReadString(LPCTSTR);
BOOL regReadBool(LPCTSTR);
void regWriteString(LPTSTR,LPCTSTR);
void regWriteBool(BOOL,LPCTSTR);
DWORD regReadDWORD(LPCTSTR);
void regWriteDWORD(DWORD,LPCTSTR);
HKEY RegOpenUser(HANDLE hUser, REGSAM desiredAccess);
BOOL RegAddKeyPermission(HKEY key, PSID user, DWORD permissions);
BOOL RegRemoveKeyPermission(HKEY key, PSID user) ;
void SetAutoCredentials(PTCHAR AutoUser, PTCHAR AutoPassword);
void SetAutoAdmin( int value );

// STRING HANDLING
char * convert_to_char(const TCHAR *from);
PWSTR DupString(PWSTR pszString);
LPTSTR TranslatedString(LPTSTR Section, LPTSTR Key);
LPWSTR wlxsdup(LPTSTR);
TCHAR * convert_to_wchar(const char *from);

// ENCRYPTION
typedef PLUGIN_API DATA_BLOB *(*DLLGETENCRYPTIONKEY)(void);                
DLLGETENCRYPTIONKEY    dllEncryptionKey;
void GetEncryptionKey(void);
void encode(DATA_BLOB *data,DATA_BLOB *result);
void decode(DATA_BLOB *data,DATA_BLOB *result);
LPTSTR encryptString(LPTSTR);
LPTSTR decryptString(LPTSTR);



// PLUGIN IMPORTS
typedef PLUGIN_API BOOL (*DLLUSERLOGIN)(LPTSTR, LPTSTR, pGinaInfo *);                
typedef PLUGIN_API BOOL (*DLLCHANGEUSERPASSWORD) (LPTSTR,LPTSTR,LPTSTR); 
typedef PLUGIN_API LPCTSTR (*DLLABOUTPLUGIN) (void);                    
typedef PLUGIN_API void (*DLLCHANGEPLUGINSETTINGS) (void);            
typedef PLUGIN_API void (*DLLLOGINHOOK) (pGinaInfo *);                       
typedef PLUGIN_API void (*DLLLOGOUTHOOK) (pGinaInfo *);                       
typedef PLUGIN_API BOOL (*DLLISREQUIRED) (void);

DLLUSERLOGIN                pluginUserLogin;
DLLCHANGEUSERPASSWORD		pluginChangeUserPassword;
DLLABOUTPLUGIN				pluginAboutPlugin;
DLLCHANGEPLUGINSETTINGS		pluginChangePluginSettings;
DLLLOGINHOOK				pluginLoginHook;
DLLLOGOUTHOOK				pluginLogoutHook;
DLLISREQUIRED               pluginIsRequired;



// WINLOGON & WLX INTERFACE STUFF

typedef BOOL (WINAPI * PFWLXNEGOTIATE)  (DWORD, DWORD *);
typedef BOOL (WINAPI * PFWLXINITIALIZE) (LPWSTR, HANDLE, PVOID, PVOID, PVOID *);
typedef VOID (WINAPI * PFWLXDISPLAYSASNOTICE) (PVOID);
typedef int  (WINAPI * PFWLXLOGGEDOUTSAS) (PVOID, DWORD, PLUID, PSID, PDWORD,
                                           PHANDLE, PWLX_MPR_NOTIFY_INFO, 
                                           PVOID *);
typedef BOOL (WINAPI * PFWLXACTIVATEUSERSHELL) (PVOID, PWSTR, PWSTR, PVOID);
typedef int  (WINAPI * PFWLXLOGGEDONSAS) (PVOID, DWORD, PVOID);
typedef VOID (WINAPI * PFWLXDISPLAYLOCKEDNOTICE) (PVOID);
typedef int  (WINAPI * PFWLXWKSTALOCKEDSAS) (PVOID, DWORD);
typedef BOOL (WINAPI * PFWLXISLOCKOK) (PVOID);
typedef BOOL (WINAPI * PFWLXISLOGOFFOK) (PVOID);
typedef VOID (WINAPI * PFWLXLOGOFF) (PVOID);
typedef VOID (WINAPI * PFWLXSHUTDOWN) (PVOID, DWORD);
typedef BOOL (WINAPI * PFWLXSCREENSAVERNOTIFY) (PVOID, BOOL *);
typedef BOOL (WINAPI * PFWLXSTARTAPPLICATION) (PVOID, PWSTR, PVOID, PWSTR);
typedef BOOL (WINAPI * PFWLXNETWORKPROVIDERLOAD) (PVOID, PWLX_MPR_NOTIFY_INFO);
typedef BOOL (WINAPI * PFWLXDISPLAYSTATUSMESSAGE) (PVOID, HDESK, DWORD, PWSTR, PWSTR);
typedef BOOL (WINAPI * PFWLXGETSTATUSMESSAGE) (PVOID, DWORD *, PWSTR, DWORD);
typedef BOOL (WINAPI * PFWLXREMOVESTATUSMESSAGE) (PVOID);
typedef BOOL (WINAPI * PFWLXGETCONSOLESWITCHCREDENTIALS) (PVOID, PVOID); // New for XP
typedef BOOL (WINAPI * PFWLXRECONNECTNOTIFY) (PVOID); // RDP needs this
typedef BOOL (WINAPI * PFWLXDISCONNECTNOTIFY) (PVOID); // RDP needs this as well

PFWLXNEGOTIATE                pfWlxNegotiate;
PFWLXINITIALIZE               pfWlxInitialize;
PFWLXDISPLAYSASNOTICE         pfWlxDisplaySASNotice;
PFWLXLOGGEDOUTSAS             pfWlxLoggedOutSAS;
PFWLXACTIVATEUSERSHELL        pfWlxActivateUserShell;
PFWLXLOGGEDONSAS              pfWlxLoggedOnSAS;
PFWLXDISPLAYLOCKEDNOTICE      pfWlxDisplayLockedNotice;
PFWLXWKSTALOCKEDSAS           pfWlxWkstaLockedSAS;
PFWLXISLOCKOK                 pfWlxIsLockOk;
PFWLXISLOGOFFOK               pfWlxIsLogoffOk;
PFWLXLOGOFF                   pfWlxLogoff;
PFWLXSHUTDOWN                 pfWlxShutdown;
PFWLXSTARTAPPLICATION         pfWlxStartApplication;
PFWLXSCREENSAVERNOTIFY        pfWlxScreenSaverNotify;
PFWLXNETWORKPROVIDERLOAD      pfWlxNetworkProviderLoad;
PFWLXDISPLAYSTATUSMESSAGE     pfWlxDisplayStatusMessage;
PFWLXGETSTATUSMESSAGE         pfWlxGetStatusMessage;
PFWLXREMOVESTATUSMESSAGE      pfWlxRemoveStatusMessage;
PFWLXGETCONSOLESWITCHCREDENTIALS	pfWlxGetConsoleSwitchCredentials; 
PFWLXRECONNECTNOTIFY                pfWlxReconnectNotify;
PFWLXDISCONNECTNOTIFY               pfWlxDisconnectNotify;


typedef DWORD (WINAPI *pThemeWait)(DWORD dwTimeout);
typedef BOOL (WINAPI *pThemeWatch)(void);

pThemeWait  fThemeWait;
pThemeWatch fThemeWatch;

typedef struct ENVIRONMENT
{
  BOOL del;	
  TCHAR *variable;
  TCHAR *value;
} ENVIRONMENT;

typedef struct _MiniAccount {
  PWSTR                   pszUsername;
  PWSTR                   pszDomain;
  PWSTR                   pszPassword;
  PWSTR                   pszLogonDate;
  PWSTR                   pszFullname;
  PWSTR                   pszUID;
  PWSTR                   pszGID;
  ENVIRONMENT            *environment;
  HANDLE                  hUserToken;
} MiniAccount, * PMiniAccount;

typedef struct _GinaContext {
  PVOID	                    msGinaContext;
  HANDLE                    hWlx;
  PWLX_DISPATCH_VERSION_1_0 pWlxFuncs;		
  MiniAccount               miniAccount;
} GinaContext, *PGinaContext;

typedef struct {
  HANDLE hWlx;
  LPWSTR station;
  PWLX_DISPATCH_VERSION_1_3 pWlxFuncs;
  HANDLE hDllInstance;
  HANDLE UserToken;
} GINA_CONTEXT, *PGINA_CONTEXT;


static PWLX_DIALOG_BOX_PARAM pfWlxDialogBoxParam = NULL;
static DLGPROC pfWlxLoggedOutSASDlgProc   = NULL;
int WINAPI MyWlxDialogBoxParam (HANDLE, HANDLE, LPWSTR, HWND, DLGPROC, LPARAM);
void HookWlxDialogBoxParam (PVOID pWinlogonFunctions);
BOOL CALLBACK MyWlxLoggedOutSASDlgProc (HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

#define IDD_WLXLOGGEDOUTSAS_DIALOG 1500 
#define IDC_MSGINA_USERNAME 1502
#define IDC_MSGINA_PASSWORD 1503

PVOID g_pWinlogon;
static DWORD g_dwVersion = WLX_VERSION_1_4;
HANDLE hGlobalWlx;
#define MSCONTEXT   (pGinaContext->msGinaContext)
#define HGLOBALWLX  (pGinaContext->hWlx)
#define PWLXFUNCS   (pGinaContext->pWlxFuncs) 
#define PMINI       (&pGinaContext->miniAccount)
#define REALGINA_PATH      TEXT("MSGINA.DLL")
#define GINASTUB_VERSION   (WLX_VERSION_1_4) // Highest version supported at
                                             // this point. Remember to modify
                                             // this as support for newer version
                                             // is added to this program.




// GLOBAL VARIABLES - CONFIGURATION AND STATE INFO
FILE *debugFile;                        // Handle to the debugFile                       
pGinaInfo settingsInfo;                 // Global pGinaInfo structure
HINSTANCE pGinahInstance;               // Handle to our DLL
HANDLE hUser;                           // Authenticated users handle, for impersonation
BOOL debugOutput;                       // Output debug info flag
BOOL keepProfiles;                      // Keep local profiles flag
BOOL allowLocking;                      // Allow users to lock the desktop flag
BOOL mapAdmin;                          // Call MapDrives() for local users as well
DWORD buttonAction;                     // What action to take for the third button
BOOL hideMotd;                          // Hide the motd flag
BOOL hidePlugin;                        // Hide plugin info flag
BOOL displayErrors;                     // Display errors even if SAM succeeds
BOOL isRequired;                        // Plugin is required (no SAM only auth)
HINSTANCE hDll;                         // Handle to the plugin DLL
BOOL forceLocal;                        // Force the local accounts password
LPTSTR defaultDomain;                   // Default domain for domain interaction
LPTSTR windowTitle;                     // Login window title
LPTSTR motdText;                        // MOTD text
LPTSTR customLogo;                      // Custom logo for login screen
LPTSTR customLocked;                    // Custom logo for locked screen
LPTSTR globalGroups;                    // Groups all users should be members of
HBITMAP logoBmp;                        // Handle to the login logo
HBITMAP locked_logoBmp;                 // Handle to the locked logo
LPTSTR sUsername;                       // Single user login username
LPTSTR sPassword;                       // Single user password
LPTSTR sDomain;                         // Single user domain
BOOL sUserLogin;                        // Single user login flag
BOOL scramblePasswords;                 // Scramble passwords on exit flag
BOOL allowSpaces;                       // Allow spaces in username flag
LPTSTR UserLoginScript;                 // Script to run at user login
BOOL disallowExpiration;                // Dont allow accounts to expire
BOOL isLoggedIn;                        // Are we in a logged in state
time_t logInTime;                       // Time that login took place
DWORD TimerID;                          // Session management timer ID
int   sessionTimeLeft;                  // Time left in the current session (in seconds)
int	defaultSessionTimeout;              // How long does the default session have (0=forever)
BOOL includeDomain;                     // Include domain in auth chain flag
BOOL manageDomain;                      // Manage accounts on the domain
BOOL acctCaching;                       // Account caching flag
DWORD acctExpireTime;                   // How long before the account expires (cache length)
BOOL useBypass;                         // Show the bypass combo box
LPTSTR pluginName;                      // Name to show for the plugin in the combo box
LPTSTR adminUsername;                   // Domain management username
LPTSTR adminPassword;                   // Domain management password
BOOL showUsernameAndAuth;               // Show username and auth type in locked dialog
BOOL showMachineName;                   // Show machine name in locked diag
BOOL enableTask;                        // Allow task manager access
BOOL enableLogout;                      // Enable logout from the locked dialog
BOOL enableRestart;                     // Enable restart from the locked dialog
BOOL disablePassword;                   // Enable password change (false) from locked dialog
BOOL trueUnlock;                        // Enable true unlock feature
BOOL disableAllEventLogs;               // Turn off all logging
BOOL hookWlx;                           // Use Hook (TS/RDP) auth method
BOOL honorLockedAccounts;               // Should we honor locked accounts
BOOL honorAutoLogin;                    // Should we honor existing auto-login data
BOOL usingAutoLogin;                    // Is auto-login data set
LPTSTR locale_file;                     // Path to the language file we will use
BOOL encryptSettings;                   // Encrypt settings option
DATA_BLOB encryptionKey;                // Key for encryption
int windowsVersion;                     // Current version of windows
wchar_t windowsBuild[2048];             // Build string
OSVERSIONINFOEX osvi;                   // Windows version information struct
BOOL useSLoginCredentials;				// True to use single user login info in the hooked wlx, false to use settingsInfo
BOOL timedOutLogoff;                    // If True, then when LoggedOnSAS is called, we just return Logoff right away
BOOL idleLogoff;                        // Idle logoff feature on, or off
int idleTime;                           // Amount of time to show the idle warning
BOOL idleInterrupted;                   // Screensaver may interrupt our idle dialog, make sure we start where were supposed to
BOOL desktopLocked;                     // Indicates whether the desktop is currently locked or not
BOOL idleLockHit;                       // User wants to lock the workstation instead of logging of on idle
BOOL idleAllowScreen;                   // Allow the user to run the screensaver from idle
BOOL idleAllowLock;                     // Allow the user to lock the wkstation from idle
BOOL idleAllowReturn;                   // Allow the user to return to the desktop from idle
BOOL idleFromLocked;                    // Allow idle dialog to show from Locked screen
BOOL pluginAutoLogin;					// Should we pass the auto-login information thru to plugins?
BOOL palShowError;						// If the plugin autologin fails, then should the error show?
LPTSTR palUsername;                     // autologin username
LPTSTR palPassword;						// autologin password
                                         
 
BOOL quickLogin;						// Show quick login button
LPTSTR quickUser;						// Quick login username
LPTSTR quickPass;						// Quick login password
LPTSTR quickCaption;					// Quick login button caption
BOOL reqDom;							// Require domain succeed
BOOL keepAccount;						// Should accounts be kept
BOOL roamingProfile;					// Should we do the roaming profile stuff....
BOOL allowLockedLogout;                 // Should anyone be able to force a logout from unlock
BOOL saveLastUsername;                  // Should we save/insert the last auth'd user
LPTSTR lastUserName;                    // Last user
int loginAttemptCount;                  // # of bad attempts

#define WINDOWS_SUCKY 0
#define WINDOWS_NT_4 1
#define WINDOWS_2000 2
#define WINDOWS_XP 3
#define WINDOWS_2003 4

#endif



                                              

                                               
                                                
